import unittest
from abcgan.hfp_model import HFP_Transformer, HFP_Critic, HFP_Generator
import torch
import abcgan.constants as const
from torch.utils.data import DataLoader, TensorDataset


def get_fake_data(bs=500, latent_dim=16):
    drivers = torch.randn((bs, const.n_driver_feat))
    bvs = torch.randn((bs, const.max_alt, const.n_bv_feat))
    src_mask = torch.ones((bs, const.max_alt), dtype=torch.bool)
    hfps = torch.randn((bs, const.n_waves, const.n_hfp_feat))
    tgt_mask = (torch.randint(10, size=(bs,)) > 4)[:, None]
    hfps[tgt_mask] = 0.0
    ds = TensorDataset(drivers, bvs, src_mask, hfps, tgt_mask)
    dl = DataLoader(ds, batch_size=bs, num_workers=0)
    dr, bv, smask, h, tmask = next(iter(dl))
    dr = dr.float()
    bv = bv.float()
    h = h.float()
    noise = torch.randn(h.shape[0] * h.shape[1],
                        latent_dim,
                        dtype=h.dtype)
    return dr, bv, smask, h, tmask, noise


class TestHFPGAN(unittest.TestCase):
    def test_hfp_generator(self):
        bs = 500
        gen_transformer = HFP_Transformer(output_b=True)
        gen_model = HFP_Generator(transformer=gen_transformer)
        dr, bv, src_mask, hfp, tgt_mask, noise = get_fake_data(bs, gen_model.latent_dim)
        fake_output, b = gen_model(dr, bv, hfp, ~src_mask, ~tgt_mask, noise)
        self.assertEqual(fake_output.shape, hfp.shape)
        self.assertEqual(b.shape[0], hfp.shape[0])
        self.assertTrue(torch.isnan(fake_output).sum() == 0)
        self.assertTrue(torch.isnan(b).sum() == 0)

    def test_critic(self):
        bs = 500
        crit_transformer = HFP_Transformer()
        crit_model = HFP_Critic(transformer=crit_transformer)
        dr, bv, src_mask, hfp, tgt_mask, noise = get_fake_data(bs, 16)
        fake_pred = crit_model(dr, bv, hfp, hfp, ~src_mask, ~tgt_mask)
        self.assertEqual(fake_pred.shape, (hfp.shape[0]*hfp.shape[1], 1))
        self.assertTrue(torch.isnan(fake_pred).sum() == 0)


if __name__ == '__main__':
    unittest.main()



import torch
import torch.nn as nn
import abcgan.constants as const
from torch.nn.init import xavier_uniform_
from torch import Tensor
from typing import Optional
from abcgan.bv_model import Generator, Critic  # noqa


class HFP_Generator(nn.Module):
    """
    Generator Class

    Parameters
    -----------
    transformer: torch.nn.Module
        the transformer model object that estimates hfp features
    n_layers: int
        number of MLP layers
    latent_dim: int
        the dimension of the input latent vector
    img_dim: int
        the dimension of the images, fitted for the dataset used, a scalar
    hidden_dim: int
        the inner dimension, a scalar
    """

    def __init__(self,
                 transformer: nn.Module,
                 n_layers=4,
                 latent_dim=16,
                 img_dim=const.n_hfp_feat,
                 hidden_dim=128, ):
        super(HFP_Generator, self).__init__()

        cond_dim = transformer.d_model + latent_dim

        self.input_args = {
            'n_layers': n_layers,
            'latent_dim': latent_dim,
            'img_dim': img_dim,
            'hidden_dim': hidden_dim,
        }

        # Build the HFP neural network
        layers = [nn.Linear(cond_dim, hidden_dim),
                  nn.ReLU(inplace=True)]
        for i in range(n_layers - 1):
            layers.append(nn.Linear(hidden_dim, hidden_dim))
            layers.append(nn.ReLU(inplace=True))
        layers.append(nn.Linear(hidden_dim, img_dim))
        self.gen_model = nn.Sequential(*layers)

        self.transformer = transformer
        self.n_layers = n_layers
        self.latent_dim = latent_dim
        self.img_dim = img_dim
        self.cond_dim = cond_dim
        self.hidden_dim = hidden_dim
        self._reset_parameters()

    def forward(self, driver_src, bv_src, hfp_tgt,
                src_key_mask=None, tgt_key_mask=None, noise=None):
        """
        Function for completing a forward pass of the generator:
        Given a noise tensor,
        returns generated images.
        Parameters
        --------------
        driver_src: torch.Tensor
            tensor of driver features from data loader (n_batch, n_dr_feat)
        bv_src: torch.Tensor
            tensor of bv featrues from data loader (n_batch, n_alt, n_bv_feat)
        hfp_tgt: torch.Tensor
            tensor of hfp featrues from data loader (n_batch, n_wave, n_hfp_feat)
        src_key_mask: torch.Tensor, optional
            mask for bv features from data loader (n_alt, n_batch)
        tgt_key_mask: torch.Tensor, optional
            mask for hfp features from data loader (n_batch, n_wave)
        noise: torch.Tensor, optional
            a noise tensor with dimensions (n_batch, latent_dim)
        """
        if src_key_mask is None:
            src_key_mask = torch.zeros(
                bv_src.shape[0], bv_src.shape[1],
                dtype=torch.bool, device=bv_src.device)
        if tgt_key_mask is None:
            tgt_key_mask = torch.zeros(
                hfp_tgt.shape[0], hfp_tgt.shape[1],
                dtype=torch.bool, device=hfp_tgt.device)

        # Get Conditioning from transformer
        _, est_hfp = self.transformer(driver_src, bv_src, hfp_tgt, src_key_mask, tgt_key_mask)
        G_b = torch.sigmoid(est_hfp[..., -1]).flatten()
        est_hfp = est_hfp[..., :-1]

        if noise is None:
            noise = torch.randn(hfp_tgt.shape[0] * hfp_tgt.shape[1],
                                self.latent_dim,
                                dtype=driver_src.dtype,
                                device=driver_src.device)

        # Add conditioning and generate estimated hfps
        G_hfp = self.gen_model(torch.cat((est_hfp.flatten(0, 1), noise), 1))
        G_hfp = G_hfp.reshape(hfp_tgt.shape[0], hfp_tgt.shape[1], hfp_tgt.shape[2])
        return G_hfp, G_b

    def _reset_parameters(self):
        """Initiate parameters in the transformer model."""
        with torch.no_grad():
            for p in self.parameters():
                if p.dim() > 1:
                    xavier_uniform_(p)


class HFP_Critic(nn.Module):
    """
    Critic Class

    Parameters
    -------------
    transformer: torch.nn.Module
        transformer for the critic
    n_layers: int
        number of layers in MLP
    img_dim: int
        the dimension of the images, fitted for the dataset used, a scalar
    hidden_dim: int
        the inner dimension, a scalar
    """

    def __init__(self,
                 transformer: nn.Module,
                 n_layers=4,
                 img_dim=const.n_hfp_feat,
                 hidden_dim=128, ):
        super(HFP_Critic, self).__init__()
        cond_dim = transformer.d_model + img_dim

        self.input_args = {
            'n_layers': n_layers,
            'img_dim': img_dim,
            'hidden_dim': hidden_dim,
        }
        # Build the neural network
        layers = [nn.Linear(cond_dim, hidden_dim),
                  nn.ReLU(inplace=True)]
        for i in range(n_layers - 1):
            layers.append(nn.Linear(hidden_dim, hidden_dim))
            layers.append(nn.ReLU(inplace=True))
        layers.append(nn.Linear(hidden_dim, 1))
        self.crit_model = nn.Sequential(*layers)

        self.n_layers = n_layers
        self.img_dim = img_dim
        self.cond_dim = cond_dim
        self.hidden_dim = hidden_dim
        self.transformer = transformer
        self._reset_parameters()

    def forward(self, dr_src, real_bv, real_hfp, hfp_feat, src_key_mask=None, tgt_key_mask=None):
        """
        Critic forward

        Parameters
        ---------------
        dr_src: torch.Tensor
            tensor of driver features from data loader (n_batch, n_dr_feat)
        real_bv: torch.Tensor
            tensor of bv features from data loader (n_batch, n_alt, n_bv_feat)
        real_hfp: torch.Tensor
            tensor of real hfp features from data loader (n_batch, n_wave, n_bv_feat)
        hfp_feat: torch.Tensor
            tensor of hfp features from data loader (n_batch, n_wave, n_bv_feat)
        src_key_mask: torch.Tensor, optional
            mask for bv features from data loader (n_batch, n_alt)
        tgt_key_mask: torch.Tensor, optional
            mask for hfp features from data loader (n_batch, n_wave)
        """
        if src_key_mask is None:
            src_key_mask = torch.zeros(
                real_bv.shape[0], real_bv.shape[1],
                dtype=torch.bool, device=real_bv.device)
        if tgt_key_mask is None:
            tgt_key_mask = torch.zeros(
                real_hfp.shape[0], real_hfp.shape[1],
                dtype=torch.bool, device=real_hfp.device)
        _, est_hfp = self.transformer(dr_src, real_bv, real_hfp, src_key_mask, tgt_key_mask)

        cond_hfp = torch.cat((est_hfp.flatten(0, 1), hfp_feat.flatten(0, 1)), 1)
        hfp_pred = self.crit_model(cond_hfp)
        return hfp_pred

    def _reset_parameters(self):
        """Initiate parameters in the transformer model."""
        with torch.no_grad():
            for p in self.parameters():
                if p.dim() > 1:
                    xavier_uniform_(p)


class HFP_Transformer(nn.Module):
    """Transformer with only the encoder
    Parameters:
        d_model: the number of expected features in the encoder/decoder inputs
        nhead: the number of heads in the multiheadattention models
        num_encoder_layers: the number of sub-encoder-layers in the encoder
        dim_feedforward: the dimension of the feedforward network model
        dropout: the dropout value
        activation: the activation function of encoder/decoder intermediate layer
    """

    def __init__(self,
                 d_dr: int = const.n_driver_feat,
                 d_bv: int = const.n_bv_feat,
                 n_alt: int = const.max_alt,
                 n_waves: int = const.n_waves,
                 d_hfp: int = const.n_hfp_feat,
                 d_model: int = 256,
                 output_b: bool = False,
                 nhead: int = 1,
                 num_encoder_layers: int = 1,
                 num_decoder_layers: int = 1,
                 dim_feedforward: int = 1024,
                 dropout: float = 0.1,
                 activation: str = "relu") -> None:
        super().__init__()
        # args captured for persistence
        self.input_args = {
            'd_dr': d_dr,
            'd_bv': d_bv,
            'n_alt': n_alt,
            'd_hfp': d_hfp,
            'n_waves': n_waves,
            'd_model': d_model,
            'output_b': output_b,
            'nhead': nhead,
            'num_encoder_layers': num_encoder_layers,
            'num_decoder_layers': num_decoder_layers,
            'dim_feedforward': dim_feedforward,
            'dropout': dropout,
            'activation': activation,
        }

        encoder_layer = nn.TransformerEncoderLayer(
            d_model, nhead, dim_feedforward, dropout, activation)
        encoder_norm = nn.LayerNorm(d_model)
        self.encoder = nn.TransformerEncoder(
            encoder_layer, num_encoder_layers, encoder_norm)

        decoder_layer = nn.TransformerDecoderLayer(
            d_model, nhead, dim_feedforward, dropout)
        decoder_norm = nn.LayerNorm(d_model)
        self.decoder = nn.TransformerDecoder(
            decoder_layer, num_encoder_layers, decoder_norm)

        self.bv_position = nn.Parameter(torch.zeros(n_alt, d_model))
        self.dr_emb = nn.Parameter(torch.zeros(d_dr, d_model))
        self.bv_emb = nn.Parameter(torch.zeros(d_bv, d_model))
        self.hfp_emb = nn.Parameter(torch.zeros(d_hfp, d_model))
        self.bv_start_token = nn.Parameter(torch.zeros(1, d_model))
        self.hfp_start_token = nn.Parameter(torch.zeros(1, d_model))
        if output_b:
            self.b_emb = nn.Parameter(torch.zeros(d_model, d_model + 1))

        self._reset_parameters()

        self.output_b = output_b
        self.d_model = d_model
        self.n_decode_layers = num_decoder_layers
        self.n_encode_layers = num_encoder_layers
        self.n_waves = n_waves
        self.d_bv = d_bv
        self.d_hfp = d_hfp
        self.nhead = nhead

        src_mask = self.generate_square_subsequent_mask(n_alt)
        self.register_buffer('src_mask', src_mask)
        tgt_mask = self.generate_square_subsequent_mask(n_waves)
        self.register_buffer('tgt_mask', tgt_mask)

    def forward(self,
                driver_src: Tensor,
                bv_src: Tensor,
                hfp_tgt: Tensor,
                src_key_padding_mask: Optional[Tensor] = None,
                tgt_key_padding_mask: Optional[Tensor] = None, ):
        """
        Take in and process masked source/target sequences.
        Args:
            driver_src: conditioning driver input (required).
            bv_src: the sequence to the encoder (required).
            hfp_tgt: the sequence to the decoder (required).
            src_key_padding_mask: the ByteTensor mask for src keys per batch
            tgt_key_padding_mask: the ByteTensor mask for tgt keys per batch
            (optional).
        Shape:
            - driver_src: :math:`(n_batch, d_dr)`.
            - bv_src: :math:`(n_batch, n_alt, d_bv)`.
            - src_key_padding_mask: :math:`(n_batch, n_alt)`.
        """
        bv_shift = torch.roll(bv_src.transpose(0, 1) @ self.bv_emb, 1, dims=0)
        bv_shift[0, :, :] = self.bv_start_token.unsqueeze(1)
        src = (driver_src.unsqueeze(0) @ self.dr_emb + bv_shift +
               self.bv_position.unsqueeze(1))

        hfp_embedded = hfp_tgt.transpose(0, 1) @ self.hfp_emb
        hfp_shift = torch.roll(hfp_embedded, 1, dims=0)
        hfp_shift[0, :, :] = self.hfp_start_token.unsqueeze(1)
        tgt = (driver_src.unsqueeze(0) @ self.dr_emb + hfp_shift)

        key_mask = src_key_padding_mask.to(self.src_mask.dtype) if src_key_padding_mask is not None else None
        encoder_output = self.encoder(src, mask=self.src_mask,
                                      src_key_padding_mask=key_mask)
        decoder_output = self.decoder(tgt, encoder_output,
                                      tgt_mask=self.tgt_mask,
                                      memory_key_padding_mask=key_mask)

        if self.output_b:
            decoder_output = (decoder_output @ self.b_emb)

        bv_output = encoder_output.transpose(0, 1)
        hfp_output = decoder_output.transpose(0, 1)
        return bv_output, hfp_output

    def generate_square_subsequent_mask(self, sz: int) -> Tensor:
        """
        Generate a square mask for the sequence. The masked positions are
        filled with float('-inf').
        Unmasked positions are filled with float(0.0).
        """
        mask = (torch.triu(torch.ones(sz, sz)) == 1).transpose(0, 1)
        mask = mask.float().masked_fill(mask == 0, float('-inf'))
        mask = mask.masked_fill(mask == 1, float(0.0))
        return mask

    def _reset_parameters(self):
        """Initiate parameters in the transformer model."""
        with torch.no_grad():
            for p in self.parameters():
                if p.dim() > 1:
                    xavier_uniform_(p)
            self.bv_position.normal_()
            self.bv_start_token.normal_()
            self.hfp_start_token.normal_()



import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
import datetime
import pandas as pd

import abcgan.constants as const

# import importlib
# importlib.reload(util)
def plot_wtec_data(data, **kwargs):
    bins = kwargs.get('bins', 100)
    tid_type = data[list(data.keys())[0]].get("tid_type", "WTEC")
    locations = list(data.keys())
    fig, ax = plt.subplots(len(const.wtec_vars) + 1, 3, figsize=(20, 20))
    ax = ax.flatten()
    for i in range(len(const.wtec_names)):
        hue = []
        for loc in locations:
            hue += [f'{loc}'] * len(data[loc]['utc'])
        sns.histplot(x=np.hstack([data[loc]['wtecs'][:, i] for loc in locations]),
                     bins=bins, ax=ax[i], element='step', hue=hue, alpha=0.5)
        ax[i].set_xlabel(f'{const.wtec_units[i]}', fontsize=15)
        ax[i].set_ylabel('Counts')
        ax[i].set_title(f'{const.wtec_names[i]}', fontsize=20)
    hue = []
    for loc in locations:
        hue += [f'{loc}'] * len(data[loc]['utc'])
    sns.histplot(x=np.hstack([[datetime.datetime.utcfromtimestamp(d) for d in data[loc]['utc']] for loc in locations]),
                 bins=bins, ax=ax[-1], hue=hue, element='step', alpha=0.8)
    ax[-1].set_xlabel(f'Date', fontsize=15)
    ax[-1].set_ylabel('Counts')
    ax[-1].set_title(f'Unix Timestamps', fontsize=20)
    plt.tight_layout(rect=[0, 0, 1, 0.95])
    plt.suptitle(f"{tid_type} Waves", fontsize=30)
    plt.show()
    plt.close(fig)


def plot_real_vs_generated_wtec(real,
                                generated,
                                hel_dists=None,
                                hist_info=None,
                                tid_type=const.wtec_default_tid_type,
                                location=const.wtec_default_location,
                                title_annot=None,
                                wtec_var="avg",
                                n_bins=100):
    if hist_info is not None:
        (r_hists, f_hists, edges) = hist_info
        if isinstance(r_hists, np.ndarray):
            r_hists = r_hists.T
            f_hists = f_hists.T
            edges = edges.T
    feat_indexes = np.array([i for i, n in enumerate(const.wtec_names) if n.find(wtec_var) > 0])
    plt.style.use('default')
    pal = sns.color_palette(n_colors=2)
    fig, ax = plt.subplots(2, 3, figsize=(15, 7))
    ax = ax.flatten()
    for i, w_index in enumerate(feat_indexes):
        x = np.hstack((real[..., w_index].flatten(), generated[..., w_index].flatten()))
        hue = ['Real'] * len(real[..., w_index].flatten()) + ['Generated'] * len(generated[..., w_index].flatten())
        data = pd.DataFrame({f'{const.wtec_units[w_index]}': x, 'Type': hue})
        if hist_info is not None:
            n_bins = len(r_hists[w_index])
        sns.histplot(data=data, x=f'{const.wtec_units[w_index]}', hue='Type', ax=ax[i],
                     element='step', bins=n_bins, legend=True, stat='density',
                     alpha=0.2, common_norm=False, palette=pal,
                     binrange=const.wtec_dict[tid_type]["meas_ranges"][w_index],
                     hue_order=['Generated', 'Real'])
        if hist_info is not None:
            if isinstance(edges, list):
                xlabel = edges[w_index]
            else:
                xlabel = edges[w_index][1:] - np.diff(edges[w_index]) / 2
            ax[i].plot(xlabel, r_hists[w_index], linewidth=2, color=pal[1], label='Actual')
            ax[i].plot(xlabel, f_hists[w_index], linewidth=2, color=pal[0], label='Generated')
        if hel_dists is not None:
            ax[i].set_title(f'{const.wtec_names[w_index]} (Hel Dist = {hel_dists[w_index]:.3f})')
        else:
            ax[i].set_title(f'{const.wtec_names[w_index]}')
    if title_annot is None:
        if hel_dists is not None:
            plt.suptitle(f'{tid_type}_{location} Model Results || Avg = {hel_dists.mean():.3f}', fontsize=20)
        else:
            plt.suptitle(f'{tid_type}_{location} Model Results', fontsize=20)
    else:
        plt.suptitle(f'{title_annot} || Avg = {hel_dists.mean():.3f}', fontsize=20)
    plt.tight_layout(rect=[0, 0, 1, 0.94])
    plt.show()
    plt.close(fig)


import unittest
import abcgan.constants as const
from abcgan.attention import collect_bv_attn_map, collect_hfp_attn_map
import numpy as np


def get_data(n):
    drivers = np.exp(np.random.normal(size=(n, const.n_driver)))
    bvs = np.exp(np.random.normal(size=(n, const.max_alt, const.n_bv)))
    return drivers, bvs


class TestAttention(unittest.TestCase):

    def test_collect_bv_attention(self):
        drivers, bvs = get_data(10)
        attn_map = collect_bv_attn_map(drivers, bvs)
        self.assertEqual(attn_map.shape, (len(drivers), const.max_alt, const.max_alt))

    def test_collect_hfp_attention(self):
        drivers, bvs = get_data(10)
        attn_map = collect_hfp_attn_map(drivers, bvs)
        self.assertEqual(attn_map.shape, (len(drivers), const.n_waves, const.max_alt))


if __name__ == "__main__":
    unittest.main(argv=['first-arg-is-ignored'], exit=False)



"""
This module supports persistence of the generator and discriminator.

It saves two files a parameters file and a configuration file.

It also supports persisting of multiple modules.

To be persistable in this way the module must have a property
containing a json serializable input dictionary as
mdl.input_args
"""
from abcgan import bv_model
from abcgan import hfp_model
from abcgan import tec_model
from abcgan import mean_estimation as me
import json
import torch
import os

dir_path = os.path.dirname(os.path.realpath(__file__))
dir_path = os.path.join(dir_path, 'models')


def fullname(inst):
    tp = type(inst)
    # check if instance is actually a type
    if isinstance(inst, type):
        tp = inst
    return tp.__module__ + '.' + tp.__name__


# list types of persistable object
types = [
    bv_model.Generator,
    bv_model.Critic,
    bv_model.Driver_Generator,
    bv_model.Driver_Critic,
    me.Transformer,
    hfp_model.HFP_Critic,
    hfp_model.HFP_Generator,
    hfp_model.HFP_Transformer,
    tec_model.WTEC_Generator,
    tec_model.WTEC_Critic
]
# register types of objects that can be persisted
type_dict = {fullname(t): t for t in types}


def persist(generator, critic, name='gan', dir_path=dir_path, train_conf=None):
    """
    Persists abcgan generator and critic modules.

    Persists both input arguments and parameters.

    Parameters:
        generator: torch.nn.Module
            module for the generator
        critic: torch.nn.Module
            module for the critic
        name: str, optional
            name of the saved configuration
        dir_path: str, optional
            default is the models directory. None assumes
            file is in local directory.
        train_conf: dict, optional
            default is None. dictionary of training parameters used

    The generator, critic and any transformers passed in as
    arguments to these must be registered in persist.py and must
    have a parameter 'input_args' that specifies their input
    arguments as a dictionary
    """
    if dir_path is not None:
        name = os.path.join(dir_path, name)
    jdict = {}
    if hasattr(generator, 'transformer'):
        jdict['gen_trans_input'] = generator.transformer.input_args
        jdict['gen_trans_type'] = fullname(generator.transformer)
    if hasattr(critic, 'transformer'):
        jdict['crit_trans_input'] = critic.transformer.input_args
        jdict['crit_trans_type'] = fullname(critic.transformer)
    jdict['gen_input'] = generator.input_args
    jdict['gen_type'] = fullname(generator)
    jdict['crit_input'] = critic.input_args
    jdict['crit_type'] = fullname(critic)
    # verify types are registered
    for k, v in jdict.items():
        if k.endswith('type'):
            if v not in type_dict:
                raise ValueError(f"Uknown module type {v}.")
    if train_conf is not None:
        jdict["train_conf"] = train_conf
    with open(name+'.json', 'w+') as f:
        json.dump(jdict, f, indent=2)
    state_dict = {}
    state_dict['gen'] = generator.state_dict()
    state_dict['crit'] = critic.state_dict()
    torch.save(state_dict, name + '.pt')


def recreate(name='gan', dir_path=dir_path):
    """
    Load a pre-trained generator and discriminator.

    Parameters
    --------------
    name: str, optional
        name of the configuration to load, as saved by
        persist. default: 'wgan_gp'
    dir_path: str, optional
        default is the models directory. None assumes
        file is in local directory.

    Returns
    -------------
    generator: torch.nn.module
        the loaded generator
    critic: torch.nn.module
        the loaded critic

    Modules must have previosuly been saved. All modules are
    loaded on the cpu, they can subsequently be moved.
    """
    if dir_path is not None:
        name = os.path.join(dir_path, name)
    with open(name + '.json', 'r') as f:
        jdict = json.load(f)
    if 'gen_trans_type' in jdict:
        trans_in = jdict['gen_trans_input']
        trans_type = type_dict[jdict['gen_trans_type']]
        jdict['gen_input']['transformer'] = trans_type(**trans_in)
    if 'crit_trans_type' in jdict:
        trans_in = jdict['crit_trans_input']
        trans_type = type_dict[jdict['crit_trans_type']]
        jdict['crit_input']['transformer'] = trans_type(**trans_in)
    gen_in = jdict['gen_input']
    gen_type = type_dict[jdict['gen_type']]
    generator = gen_type(**gen_in)
    crit_in = jdict['crit_input']
    crit_type = type_dict[jdict['crit_type']]
    critic = crit_type(**crit_in)
    state_dict = torch.load(name + '.pt',
                            map_location=torch.device('cpu'))
    generator.load_state_dict(state_dict['gen'])
    critic.load_state_dict(state_dict['crit'])

    return generator, critic



import torch
import numpy as np
from typing import Union, Tuple
from scipy import stats

import abcgan.constants as const
import abcgan.transforms as trans


def get_clusters(drivers: np.ndarray, n_clusters: int = 10):
    """
    Returns cluster indicating conditional subsets based on driver values

    Parameters
    ----------------
    drivers: np.ndarray
        array of drivers
    n_clusters: int
        number of clusters subsets to create for each driver feature

    Returns
    -------------
    clusters:
        array of clusters for each sample and input driver feat
    """
    clusters = np.zeros_like(drivers, dtype=int)
    for d in range(drivers.shape[-1]):
        steps = np.ones(n_clusters, dtype=int) * len(drivers) // n_clusters
        for i in range(len(drivers) % n_clusters):
            steps[i] += 1
        idxs = np.hstack((np.array([0]), np.cumsum(steps)))
        driver_sort_idxs = np.argsort(drivers[:, d])
        for i in range(1, len(idxs)):
            clusters[driver_sort_idxs[idxs[i - 1]: idxs[i]], d] = i - 1
    return clusters.astype(int)


def conditional_wtec_scores(drivers: np.ndarray,
                            real_wtecs: np.ndarray,
                            generated_wtecs: np.ndarray,
                            n_clusters: int = 5,
                            tid_type=const.wtec_default_tid_type,
                            return_hist_info: bool = False,
                            **hellinger_kwargs):
    """
    Returns the hellinger distance scores across various conditional subsets derived
    from the geo physical driver values

    Parameters
    ----------------
    drivers: np.ndarray
        drivers to condition subsets with (Not Z-Scaled)
    real_wtecs: np.ndarray
        array of tec wave parameters (Not Z-Scaled)
    generated_wtecs: np.ndarray
        array of generated tec wave parameters (Not Z-Scaled)
    n_clusters: int
        number of conditional subsets to create for each driver feature
    tid_type: str
       TID type for scaling
    return_hist_info:
        set to return histograms and edges
    hellinger_kwargs: dict
        hellinger distance calculation arguments

    Returns
    -------------
    cond_dist:
        the hellinger distance for each cluster, driver and wtec feature
         (n_drivers or n_clusters x n_wtec)
    hist_info:
        additional histogram data that was used to calculate hellinger dist.
        Info includes to real and fake histograms and there shared bin edges.
    """
    if n_clusters < 1:
        raise ValueError(f'n_clusters = {n_clusters} is invalid. Must be greater than zero.')

    if hellinger_kwargs.get("z_scale") is not None:
        hellinger_kwargs.pop("z_scale")
    if hellinger_kwargs.get("z_scale_inputs") is not None:
        hellinger_kwargs.pop("z_scale_inputs")
    clusters = get_clusters(drivers, n_clusters=n_clusters)
    cond_dist = np.zeros((drivers.shape[-1], n_clusters, real_wtecs.shape[-1]))
    hist_info = [[] for i in range(drivers.shape[-1])]

    for i in range(drivers.shape[-1]):
        for j in range(n_clusters):
            c_mask = clusters[:, i] == j
            cond_dist[i, j, :], info = hellinger_scores_wtec(real_wtecs[c_mask],
                                                             generated_wtecs[c_mask],
                                                             z_scale=False, z_scale_inputs=False,
                                                             tid_type=tid_type,
                                                             return_hist_info=True,
                                                             **hellinger_kwargs)
            hist_info[i].append(info)
    if return_hist_info:
        return cond_dist, hist_info
    else:
        return cond_dist


def hellinger_score_kde(ds1: np.ndarray,
                        ds2: np.ndarray,
                        binranges: Union[None, tuple] = None,
                        bins: Union[None, str, int] = 'auto',
                        return_hist_info: bool = False,):
    """
    Returns the hellinger distance score that measures the similarity between two
    distributions. Uses kde to generate pdfs instead of histograms. Produces good
    results without having to fine tune bin size or averaging filter parameters,
    however takes longer to compute.

    Parameters
    ----------------
    ds1:
        array of first dataset
    ds2:
        array of second dataset
    binranges:
        ranges to include in the histograms
    bins:
        the number of bins to use, default is auto
    return_hist_info:
        set to return histograms and edges
    Returns
    -------------
    dist:
        the hellinger distance (n_alts or n_waves x n_feats)
    hist_info:
        additional histogram data that was used to calculate hellinger dist.
        Info includes to real and fake histograms and there shared bin edges.
    """

    if bins is None:
        bins = 'auto'
    if binranges is None:
        hist_args = {'bins': bins, 'density': True}
    else:
        hist_args = {'bins': bins, 'range': binranges, 'density': True}

    if len(ds1) > len(ds2):
        hist, edg = np.histogram(ds1, **hist_args)
    else:
        hist, edg = np.histogram(ds2, **hist_args)

    kernel_1 = stats.gaussian_kde(ds1)
    kernel_2 = stats.gaussian_kde(ds2)

    x_pos = edg[1:] - np.diff(edg) / 2
    h1 = kernel_1(x_pos)
    h2 = kernel_2(x_pos)
    a1 = h1 * np.diff(edg)
    a2 = h2 * np.diff(edg)
    dist = (1 / np.sqrt(2)) * np.sqrt(np.sum((np.sqrt(a1) - np.sqrt(a2)) ** 2))

    if return_hist_info:
        return dist, (h1, h2, x_pos)
    else:
        return dist


def hellinger_dist(real, fake, mask=None, bins=None, filter_length=None, z_ranges=None):
    """
    Returns the hellinger distance score between two distributions

    Parameters
    ----------------
    real:
        tensor of real values for a particular alt and bv feat
    fake:
        tensor of generated values for a particular alt and bv feat
    mask:
        mask for bv features
    bins: int
        number of bins to use in histogram calculations
        (If None # of bins will be calculated based on number of samples)
    filter_length: int
        averaging filter length to smooth out noise in histograms
        (If None filter length will be calculated based on number of samples)
        bins: int
        number of bins to use in histogram calculations
        (If None # of bins will be calculated based on number of samples)
    z_ranges: tuple
        upper and lower bounds to evaluate histograms
    Returns
    -------------
    penalty:
        the hellinger distance
    """

    if bins is None:
        bins = max(15, int((len(real)**const.bin_exp)))
    if filter_length is None:
        filter_length = max(2, int(len(real)**const.filter_exp))

    if mask is None:
        r = real.flatten().detach().cpu().numpy()
        f = fake.flatten().detach().cpu().numpy()
    else:
        r = real[mask].flatten().detach().cpu().numpy()
        f = fake[mask].flatten().detach().cpu().numpy()

    if z_ranges is None:
        binranges = (min(np.nanmin(r), np.nanmin(f)),
                     max(np.nanmax(r), np.nanmax(f)))
    else:
        binranges = (max(z_ranges[0], min(np.nanmin(r), np.nanmin(f))),
                     min(z_ranges[1], max(np.nanmax(r), np.nanmax(f))))
        if (f < binranges[0]).all() or (f > binranges[1]).all():
            binranges = (min(np.nanmin(r), np.nanmin(f)), max(np.nanmax(r), np.nanmax(f)))

    args = {'bins': bins, 'range': binranges, 'density': True}
    r_hist, r_edges = np.histogram(r, **args)
    f_hist, f_edges = np.histogram(f, **args)
    if filter_length:
        r_hist = np.convolve(r_hist, np.ones(filter_length), mode='same') / filter_length
        f_hist = np.convolve(f_hist, np.ones(filter_length), mode='same') / filter_length
    dist = (1 / np.sqrt(2)) * np.sqrt(np.sum((np.sqrt(r_hist * np.diff(r_edges)) -
                                              np.sqrt(f_hist * np.diff(f_edges))) ** 2))
    return dist


def hellinger_scores_bv(real: np.ndarray,
                        fake: np.ndarray,
                        mask: Union[None, np.ndarray] = None,
                        bins: Union[None, int] = None,
                        filter_length: Union[None, int] = None,
                        return_hist_info: bool = False,
                        z_scale: bool = True,
                        z_scale_input: bool = False,
                        bv_type: str = 'radar'):
    """
    Returns the hellinger distance score that measures how similarity between
    real and generated background variable profiles.

    Parameters
    ----------------
    real: np.ndarray
        tensor of real values for a particular alt and bv feat
    fake: np.ndarray
        tensor of generated values for a particular alt and bv feat
    bins: int
        number of bins to use in histogram calculations
        (If None # of bins will be calculated based on number of samples)
    filter_length: int
        averaging filter length to smooth out noise in histograms
        (If None filter length will be calculated based on number of samples)
    return_hist_info: bool
        set to have function return the histograms and bin edges used which
        were used to calculate the hellinger distance metric.
    z_scale: bool
        used z-scaled values when calculating hellinger distance (recommended)
    z_scale_input: bool
        Set if you are inputting bvs that are
        already z-scaled
    bv_type:
        type of data (radar or lidar)
    Returns
    -------------
    dist:
        the hellinger distance (n_alts x n_feats)
    hist_info
        additional histogram data that was used to calculate hellinger dist.
        Info includes to real and fake histograms and there shared bin edges.
    """
    if mask is None:
        mask = np.ones((real.shape[0], real.shape[1]), dtype=bool)

    if bins is None:
        bins = max(15, int((real.shape[0])**const.bin_exp))
    if filter_length is None:
        filter_length = max(2, int(len(real)**const.filter_exp))

    dists = np.zeros((real.shape[1], real.shape[2]))
    r_hists = np.zeros((bins, real.shape[1], real.shape[2]))
    f_hists = np.zeros((bins, real.shape[1], real.shape[2]))
    edges = np.zeros((bins + 1, real.shape[1], real.shape[2]))

    for i in range(real.shape[1]):
        for j in range(real.shape[2]):
            if z_scale:
                if z_scale_input:
                    r = real[mask[:, i], i, j]
                    f = fake[mask[:, i], i, j]
                else:
                    r = trans.scale_bv(real, bv_type)[0][mask[:, i], i, j]
                    f = trans.scale_bv(fake, bv_type)[0][mask[:, i], i, j]
                binranges = (max(const.bv_z_ranges[j][0], min(np.nanmin(r), np.nanmin(f))),
                             min(const.bv_z_ranges[j][1], max(np.nanmax(r), np.nanmax(f))))
                args = {'bins': bins, 'range': binranges, 'density': True}
            else:
                if z_scale_input:
                    r = trans.get_bv(real, bv_type)[mask[:, i], i, j]
                    f = trans.get_bv(fake, bv_type)[mask[:, i], i, j]
                else:
                    r = real[mask[:, i], i, j]
                    f = fake[mask[:, i], i, j]
                binranges = (max(const.bv_thresholds[j, 0], min(np.nanmin(r), np.nanmin(f))),
                             min(const.bv_thresholds[j, 1], max(np.nanmax(r), np.nanmax(f))))
                args = {'bins': bins, 'range': binranges, 'density': True}
            r_hist, edg = np.histogram(r, **args)
            f_hist, edg = np.histogram(f, **args)
            if filter_length:
                r_hist = np.convolve(r_hist, np.ones(filter_length), mode='same') / filter_length
                f_hist = np.convolve(f_hist, np.ones(filter_length), mode='same') / filter_length
            r_area = r_hist * np.diff(edg)
            f_area = f_hist * np.diff(edg)
            dists[i, j] = (1 / np.sqrt(2)) * np.sqrt(np.sum((np.sqrt(r_area) - np.sqrt(f_area)) ** 2))
            r_hists[:, i, j] = r_hist
            f_hists[:, i, j] = f_hist
            edges[:, i, j] = edg

    if return_hist_info:
        return dists, (r_hists, f_hists, edges)
    else:
        return dists


def hellinger_scores_hfp(real: np.ndarray,
                         fake: np.ndarray,
                         r_mask: Union[None, np.ndarray] = None,
                         f_mask: Union[None, np.ndarray] = None,
                         n_bins: Union[None, tuple, int] = None,
                         filter_length: Union[None, int] = None,
                         return_hist_info: bool = False,
                         z_scale: bool = True,
                         z_scale_input: bool = False,):
    """
    Returns the hellinger distance score that measures the similarity between
    real and generated background variable profiles.

    Parameters
    ----------------
    real:
        tensor of real values for a particular alt and bv feat
    fake:
        tensor of generated values for a particular alt and bv feat
    n_bins:
        tensor of real values for a particular alt and bv feat
    filter_length:
        averaging filter length to smooth out histograms
    z_scale: bool
        used z-scaled values (recommended)
    z_scale_input: bool
        Set if you are inputting hfps that are
        already z-scaled
    return_hist_info: bool
        set to have function return the real hist,
        fake hist, and bin edges used in calculation
    Returns
    -------------
    dist:
        the hellinger distance (n_alts or n_waves x n_feats)
    hist_info:
        additional histogram data that was used to calculate hellinger dist.
        Info includes to real and fake histograms and there shared bin edges.
    """
    if r_mask is None:
        r_mask = np.ones((real.shape[0], real.shape[1]), dtype=bool)
    if f_mask is None:
        f_mask = np.ones((fake.shape[0], fake.shape[1]), dtype=bool)

    if n_bins is None:
        n_bins = max(15, int(r_mask.sum() ** const.bin_exp))
    if filter_length is None:
        filter_length = max(2, int(r_mask.sum() ** const.filter_exp))

    dists = np.zeros((real.shape[1], real.shape[2]))
    r_hists = np.zeros((n_bins, real.shape[1], real.shape[2]))
    f_hists = np.zeros((n_bins, real.shape[1], real.shape[2]))
    edges = np.zeros((n_bins + 1, real.shape[1], real.shape[2]))

    for i in range(real.shape[1]):
        for j in range(real.shape[2]):
            if z_scale:
                if z_scale_input:
                    r = real[r_mask[:, i], i, j]
                    f = fake[f_mask[:, i], i, j]
                else:
                    r = trans.scale_hfp(real)[0][r_mask[:, i], i, j]
                    f = trans.scale_hfp(fake)[0][f_mask[:, i], i, j]
                binranges = (max(const.hfp_z_ranges[j][0], min(np.nanmin(r), np.nanmin(f))),
                             min(const.hfp_z_ranges[j][1], max(np.nanmax(r), np.nanmax(f))))
                args = {'bins': n_bins, 'range': binranges, 'density': True}
            else:
                if z_scale_input:
                    r = trans.get_hfp(real)[r_mask[:, i], i, j]
                    f = trans.get_hfp(fake)[f_mask[:, i], i, j]
                else:
                    r = real[r_mask[:, i], i, j]
                    f = fake[f_mask[:, i], i, j]
                binranges = (max(const.hfp_meas_ranges[j][0], min(np.nanmin(r), np.nanmin(f))),
                             min(const.hfp_meas_ranges[j][1], max(np.nanmax(r), np.nanmax(f))))
                args = {'bins': n_bins, 'range': binranges, 'density': True}
            r_hist, edg = np.histogram(r, **args)
            f_hist, edg = np.histogram(f, **args)
            if filter_length:
                r_hist = np.convolve(r_hist, np.ones(filter_length), mode='same') / filter_length
                f_hist = np.convolve(f_hist, np.ones(filter_length), mode='same') / filter_length
            r_area = r_hist * np.diff(edg)
            f_area = f_hist * np.diff(edg)
            dists[i, j] = (1 / np.sqrt(2)) * np.sqrt(np.sum((np.sqrt(r_area) -
                                                             np.sqrt(f_area)) ** 2))
            r_hists[:, i, j] = r_hist
            f_hists[:, i, j] = f_hist
            edges[:, i, j] = edg

    if return_hist_info:
        return dists, (r_hists, f_hists, edges)
    else:
        return dists


def hellinger_scores_wtec(real: Union[np.ndarray, torch.Tensor],
                          fake: Union[np.ndarray, torch.Tensor],
                          n_bins: Union[None, int] = None,
                          filter_length: Union[None, int] = None,
                          z_scale: bool = False,
                          z_scale_inputs: bool = False,
                          tid_type: str = 'SSTIDs',
                          use_kde: bool = False,
                          return_hist_info: bool = False,):
    """
    Returns the hellinger distance score that measures the similarity between
    real and generated tec wave.

    Parameters
    ----------------
    real:
        array of real tec waves
    fake:
        array of fake/generated tec waves
    n_bins:
        number of bins to use during hellinger score calculation
    filter_length:
        averaging filter length to smooth out histograms
    z_scale: bool
        used z-scaled values (recommended)
    z_scale_inputs: bool
        Set if you are inputting hfps that are
        already z-scaled
    tid_type: str
        specify dataset type for z-scaling
    use_kde: bool
        use kernel density estimation to generate PDFs
    return_hist_info: bool
        set to have function return the real hist,
        fake hist, and bin edges used in calculation
    Returns
    -------------
    dist:
        the hellinger distance (1 x n_feats)
    hist_info:
        additional histogram data that was used to calculate hellinger dist.
        Info includes to real and fake histograms and there shared bin edges.
    """

    if n_bins is None:
        n_bins = max(15, int(len(real) ** const.bin_exp))
    if filter_length is None:
        filter_length = max(2, int(len(real) ** const.filter_exp))

    dists = np.zeros(real.shape[1])
    if use_kde:
        r_hists, f_hists, edges = [], [], []
    else:
        r_hists = np.zeros((n_bins, real.shape[1]))
        f_hists = np.zeros((n_bins, real.shape[1]))
        edges = np.zeros((n_bins + 1, real.shape[1]))

    for i in range(real.shape[1]):
        if z_scale:
            if z_scale_inputs:
                r = real[:, i]
                f = fake[:, i]
            else:
                r = trans.scale_wtec(real, tid_type=tid_type)[0][:, i]
                f = trans.scale_wtec(fake, tid_type=tid_type)[0][:, i]
            binranges = (max(const.wtec_dict[tid_type]['z_ranges'][i][0], min(np.nanmin(r), np.nanmin(f))),
                         min(const.wtec_dict[tid_type]['z_ranges'][i][1], max(np.nanmax(r), np.nanmax(f))))
            if (f < binranges[0]).all() or (f > binranges[1]).all():
                binranges = (min(np.nanmin(r), np.nanmin(f)), max(np.nanmax(r), np.nanmax(f)))
            args = {'bins': n_bins, 'range': binranges, 'density': True}
        else:
            if z_scale_inputs:
                r = trans.get_wtec(real, tid_type=tid_type)[:, i]
                f = trans.get_wtec(fake, tid_type=tid_type)[:, i]
            else:
                r = real[:, i]
                f = fake[:, i]
            binranges = (max(const.wtec_dict[tid_type]['meas_ranges'][i][0], min(np.nanmin(r), np.nanmin(f))),
                         min(const.wtec_dict[tid_type]['meas_ranges'][i][1], max(np.nanmax(r), np.nanmax(f))))
            if (f < binranges[0]).all() or (f > binranges[1]).all():
                binranges = (min(np.nanmin(r), np.nanmin(f)), max(np.nanmax(r), np.nanmax(f)))
            args = {'bins': n_bins, 'range': binranges, 'density': True}

        if use_kde:
            dists[i], (r_hist, f_hist, edg) = hellinger_score_kde(r, f, binranges=binranges, return_hist_info=True)
            r_hists.append(r_hist)
            f_hists.append(f_hist)
            edges.append(edg)
        else:
            r_hist, edg = np.histogram(r, **args)
            f_hist, edg = np.histogram(f, **args)
            if filter_length:
                r_hist = np.convolve(r_hist, np.ones(filter_length), mode='same') / filter_length
                f_hist = np.convolve(f_hist, np.ones(filter_length), mode='same') / filter_length
            r_area = r_hist * np.diff(edg)
            f_area = f_hist * np.diff(edg)
            dists[i] = (1 / np.sqrt(2)) * np.sqrt(np.sum((np.sqrt(r_area) -
                                                          np.sqrt(f_area)) ** 2))
            r_hists[:, i] = r_hist
            f_hists[:, i] = f_hist
            edges[:, i] = edg

    if return_hist_info:
        return dists, (r_hists, f_hists, edges)
    else:
        return dists



import unittest
from abcgan.mean_estimation import Transformer
from abcgan.bv_model import Generator, Critic, Driver_Generator, Driver_Critic
from torch.utils.data import DataLoader, TensorDataset
import abcgan.constants as const
import abcgan.transforms as trans
import numpy as np
import torch


def fake_drivers(n):
    return np.exp(np.random.normal(size=(n, const.n_driver)))


def fake_bvs(n):
    bvs = [np.random.uniform(low=const.bv_thresholds[i, 0],
                             high=const.bv_thresholds[i, 1],
                             size=(n, const.max_alt)) for i in range(const.n_bv)]
    return np.stack(bvs, axis=-1)


def fake_alt_mask(n):
    alt_mask = np.ones((n, const.max_alt), dtype=bool)
    end_alt = np.random.randint(low=2, high=const.max_alt, size=(n,))
    for i in range(n):
        alt_mask[i, end_alt[i]:] = False
    return alt_mask


def get_fake_bv_data(bs=500, latent_dim=16):
    drivers = torch.tensor(trans.scale_driver(fake_drivers(2 * bs)), dtype=torch.float)
    bvs = torch.tensor(trans.scale_bv(fake_bvs(2 * bs))[0], dtype=torch.float)
    alt_mask = torch.tensor(fake_alt_mask(2 * bs), dtype=torch.bool)
    ds = TensorDataset(drivers, bvs, alt_mask)
    dl = DataLoader(ds, batch_size=bs, num_workers=0)
    dr, bv, mask = next(iter(dl))
    dr = dr.float()
    bv = bv.float()
    mask = ~mask
    noise = torch.randn(bs * const.max_alt, latent_dim).float()
    return dr, bv, mask, noise


class TestBVModel(unittest.TestCase):
    def test_generator(self):
        bs = 500
        gen_transformer = Transformer()
        gen_model = Generator(transformer=gen_transformer)
        dr, bv, mask, noise = get_fake_bv_data(bs, gen_model.latent_dim)
        fake_output = gen_model(dr, bv, ~mask, noise)
        self.assertEqual(fake_output.shape, bv.shape)

    def test_critic(self):
        bs = 500
        crit_transformer = Transformer()
        crit_model = Critic(transformer=crit_transformer)
        dr, fake_input, mask, _ = get_fake_bv_data(bs)
        fake_pred = crit_model(fake_input, dr, fake_input, ~mask)
        self.assertEqual(fake_pred.shape, (fake_input.shape[0]*fake_input.shape[1], 1))

    def test_dr_generator(self):
        bs = 500
        gen_model = Driver_Generator()
        dr, _, _, _ = get_fake_bv_data(bs, gen_model.latent_dim)
        fake_output = gen_model(dr)
        self.assertEqual(fake_output.shape, dr.shape)

    def test_dr_critic(self):
        bs = 500
        crit_model = Driver_Critic()
        dr, _, _, _ = get_fake_bv_data(bs)
        fake_pred = crit_model(dr, dr)
        self.assertEqual(fake_pred.shape, (dr.shape[0], 1))


if __name__ == '__main__':
    unittest.main(argv=['first-arg-is-ignored'], exit=False)



import numpy as np
import abcgan.transforms as trans
from scipy.special import logsumexp
from numpy import linalg as LA
import abcgan.constants as const


def marginal_anomaly_estimation(sampled_feats: np.ndarray, feats: np.ndarray, alpha: float = 2.0):
    """
        compute an anomaly scores from the set of generated features
        using marginal distribution based logsumexp computation

        Parameters
        -------------
        sampled_feats: np.ndarray
            sampled z-scaled features for each input feature
        feats: np.ndarray
            input feature broadcast to match sampled_feat dim
        alpha: float
            scalar parameter for sigma (lower alpha --> finner resolution)
        Returns
        -------------
        anomalies: np.ndarray
             marginal anomaly scores (feat.shape)
        """

    std = np.std(sampled_feats, axis=0)[None, ...]
    sigma = std / len(sampled_feats) * alpha

    a = -0.5 * ((sampled_feats - feats) / sigma) ** 2
    b = 1 / (sigma * np.sqrt(2 * np.pi))
    anomalies = logsumexp(a=a, b=b, axis=0)
    return anomalies


def joint_anomaly_estimation(sampled_feats: np.ndarray, feats: np.ndarray, alpha: float = 1.0):
    """
        compute an anomaly scores from the set of generated features
        using joint distribution based logsumexp computation

        Parameters
        -------------
        sampled_feats: np.ndarray
            sampled z-scaled features for each input feature
        feats: np.ndarray
            input feature broadcast to match sampled_feat dim
        alpha: float
            scalar parameter for sigma (lower alpha --> finner resolution)
        Returns
        -------------
        anomalies: np.ndarray
            joint anomaly scores (feat.shape[:-1])
        """
    (n, n_feats) = sampled_feats.shape[0], sampled_feats.shape[-1]

    sigma = (np.prod(np.std(sampled_feats, axis=0), axis=-1) / n) ** (1 / n_feats) * alpha
    a = -0.5 * LA.norm(sampled_feats - feats, axis=-1) ** 2 / sigma ** 2
    b = 1 / (sigma * np.sqrt(2 * np.pi))
    anomalies = logsumexp(a=a, b=b, axis=0)

    return anomalies


def anomaly_score_bv(bvs, gen_bvs, method: str = 'joint', alpha: float = 2.0):
    """
    returns unbounded anomaly scores for a background profiles
    given set of generated samples. Low scores represent anomalous events.

    Parameters
    -------------
    bvs: np.ndarray
         (n_samples x n_alt x n_bv) real background profiles
    gen_bvs: np.ndarray
        (n_samples x n_repeat x n_alt x n_feat) n_repeat generated
         set of generated background profiles for each input sample
    method: str, optional
        'joint': estimates a single anomaly score for each altitude bin using joint distribution
        'marginal': estimates anomaly scores at each alt for each bv feature using marginal distributions
    alpha: float
        scalar parameter for sigma (lower alpha --> finner resolution)
    Returns
    -------------
    anomalies: np.ndarray
        (n_samples x n_alt)  output of anomaly scores if joint
        (n_samples x n_alt x n_feat) output of anomaly scores if marginal

    """
    if method == 'joint':
        anomalies = np.zeros(bvs.shape[:-1])
    elif method == 'marginal':
        anomalies = np.zeros_like(bvs)
    else:
        raise ValueError(f"{method} is an invalid method."
                         f" Pleas use 'marginal' or 'joint' method")
    n_alts = bvs.shape[-2]
    for i in range(len(bvs)):
        if bvs.shape[-1] == const.n_bv_feat:
            G_feats = trans.scale_bv(gen_bvs[i, ...])[0][:, :n_alts]
            real_feat = trans.scale_bv(bvs[[i], ...])[0][:, :n_alts]
        elif bvs.shape[-1] == const.n_lidar_bv_feat:
            G_feats = trans.scale_bv(gen_bvs[i, ...], bv_type='lidar')[0][:, :n_alts]
            real_feat = trans.scale_bv(bvs[[i], ...], bv_type='lidar')[0][:, :n_alts]
        else:
            raise ValueError(f"{bvs.shape[-1]} is an invalid number of input features. "
                             f"Must input {const.n_bv} for Radar BVs or "
                             f"{const.n_lidar_bv} for Lidar BVs.")
        if method == 'joint':
            anomalies[i, :] = joint_anomaly_estimation(G_feats, real_feat, alpha=alpha)
        else:
            anomalies[i, :, :] = marginal_anomaly_estimation(G_feats, real_feat, alpha=alpha)

    return anomalies


def anomaly_score_hfp(hfps: np.ndarray, gen_hfps: np.ndarray, method: str = 'joint', alpha: float = 2.0):
    """
    returns unbounded anomaly scores for a HFP waves
    given set of generated samples. Low scores represent anomalous events.

    Parameters
    -------------
    hfps: np.ndarray
         (n_samples x n_waves x n_bv) real HFPs
    gen_hfps: np.ndarray
        (n_samples x n_repeat x n_waves x n_feat) n_repeat generated
         set of generated hfp waves for each input sample
    method: str, optional
        'joint': estimates a single anomaly score for each wave using joint distribution
        'marginal': estimates anomaly scores on each hfp feature for each wave using marginal distributions
    alpha: float
        scalar parameter for sigma (lower alpha --> finner resolution)
    Returns
    -------------
    anomalies: np.ndarray
        (n_samples x n_waves)  output of anomaly scores if joint
        (n_samples x n_waves x n_feat) output of anomaly scores if marginal

    """
    if method == 'joint':
        anomalies = np.zeros(hfps.shape[:-1])
    elif method == 'marginal':
        anomalies = np.zeros_like(hfps)
    else:
        raise ValueError(f"{method} is an invalid method."
                         f" Pleas use 'marginal' or 'joint' method")

    for i in range(len(hfps)):
        if hfps.shape[-1] == const.n_hfp_feat:
            G_feats = trans.scale_hfp(gen_hfps[i, ...])[0]
            real_feat = trans.scale_hfp(hfps[[i], ...])[0]
        else:
            raise ValueError(f"{hfps.shape[-1]} is an invalid number of input features. "
                             f"Must input {const.n_hfp} for HFP Waves ")
        if method == 'joint':
            anomalies[i, :] = joint_anomaly_estimation(G_feats, real_feat, alpha=alpha)
        else:
            anomalies[i, :, :] = marginal_anomaly_estimation(G_feats, real_feat, alpha=alpha)

    return anomalies


def anomaly_score_wtec(wtecs: np.ndarray,
                       gen_wtecs: np.ndarray,
                       method: str = 'joint',
                       alpha: float = 2.0,
                       tid_type: str = const.wtec_default_tid_type):
    """
    returns unbounded anomaly scores for TEC wave parameters
    given set of generated TEC waves. Low scores represent anomalous events.

    Parameters
    -------------
    wtecs: np.ndarray
         (n_samples x n_wtec) real background profiles
    gen_wtecs: np.ndarray
        (n_samples x n_repeat x n_wtec) n_repeat generated
         set of generated tec waves for each input sample
    method: str, optional
        'marginal': estimates anomaly scores on each tec feature using marginal distributions
        'joint': estimates anomaly score of each tec wave using joint distribution
    alpha: float
        scalar parameter for sigma (lower alpha --> finner resolution)
    tid_type: str
        specify dataset type for z-scaling
    Returns
    -------------
    anomalies: np.ndarray
        (n_samples)  output of anomaly scores if joint
        (n_samples x n_feat) output of anomaly scores if marginal

    """
    if method == 'joint':
        anomalies = np.zeros(wtecs.shape[:-1])
    elif method == 'marginal':
        anomalies = np.zeros_like(wtecs)
    else:
        raise ValueError(f"{method} is an invalid method."
                         f" Pleas use 'marginal' or 'joint' method")

    for i in range(len(wtecs)):
        if wtecs.shape[-1] == const.n_wtec:
            G_feats = trans.scale_wtec(gen_wtecs[i, ...], tid_type=tid_type)[0]
            real_feat = trans.scale_wtec(wtecs[[i], ...], tid_type=tid_type)[0]
        else:
            raise ValueError(f"{wtecs.shape[-1]} is an invalid number of input features. "
                             f"Must input {const.n_wtec} for TEC Waves ")
        if method == 'joint':
            anomalies[i, ...] = joint_anomaly_estimation(G_feats, real_feat, alpha=alpha)
        else:
            anomalies[i, ...] = marginal_anomaly_estimation(G_feats, real_feat, alpha=alpha)

    return anomalies



import torch
import numpy as np
from typing import List

import abcgan.constants as const
import abcgan.transforms as trans
from abcgan import persist
from abcgan.mask import mask_altitude


def get_encoder_attn(layers, src, src_mask, src_key_mask):
    """
    Collects attention masks from each encoder layer in the transformer

    Parameters
    -------------
    layers: list
        list of pytorch encoder layers from transformer
    src: torch.tensor
        embedded source tensor
    src_mask: torch.tensor
        constant source key mask from transformer
    src_key_mask: torch.tensor
        mask for missing source data

    Returns
    -------------
    samples: torch.tensor
        n_layers x n_samples x n_alt x n_alts output of attention mask or weights
        for each encoder layer in the transformer
    """
    output = src.clone()
    attention_masks = torch.zeros((len(layers), src.shape[1], src.shape[0], src.shape[0]))

    s2, att_mask = layers[0].self_attn(output, output, output, attn_mask=src_mask,
                                       key_padding_mask=src_key_mask)
    attention_masks[0, ...] = att_mask
    s1 = src + layers[0].dropout1(s2)
    s1 = layers[0].norm1(s1)
    s2 = layers[0].linear2(layers[0].dropout(layers[0].activation(layers[0].linear1(s1))))
    s1 = s1 + layers[0].dropout2(s2)
    s1 = layers[0].norm2(s1)
    for i in range(1, len(layers)):
        s2, att_mask = layers[i].self_attn(s1, s1, s1, attn_mask=src_mask,
                                           key_padding_mask=src_key_mask)
        attention_masks[i, ...] = att_mask
        s1 = s1 + layers[i].dropout1(s2)
        s1 = layers[i].norm1(s1)
        s2 = layers[i].linear2(layers[i].dropout(layers[i].activation(layers[i].linear1(s1))))
        s1 = s1 + layers[i].dropout2(s2)
        s1 = layers[i].norm2(s1)

    return attention_masks


def get_decoder_attn(transformer, tgt, memory, key_padding_mask):

    tgt = tgt.clone()
    layers = transformer.decoder.layers
    attn_map = torch.zeros(len(layers), tgt.shape[1],
                           const.n_waves, memory.shape[0])

    for i in range(len(layers)):
        tgt2 = layers[i].self_attn(tgt, tgt, tgt, attn_mask=transformer.tgt_mask)[0]
        tgt = tgt + layers[i].dropout1(tgt2)
        tgt = layers[i].norm1(tgt)
        tgt2, attn_map[i, ...] = layers[i].multihead_attn(tgt, memory, memory,
                                                          key_padding_mask=key_padding_mask)
        tgt = tgt + layers[i].dropout2(tgt2)
        tgt = layers[i].norm2(tgt)
        tgt2 = layers[i].linear2(layers[i].dropout(layers[i].activation(layers[i].linear1(tgt))))
        tgt = tgt + layers[i].dropout3(tgt2)
        tgt = layers[i].norm3(tgt)

    return attn_map


def collect_hfp_attn_map(drivers: np.ndarray,
                         bvs: np.ndarray,
                         driver_names: List[str] = const.driver_names,
                         model: str = 'hfp_gan',
                         bv_type: str = 'radar'):
    """
    function to collect attention map from the HFP GAN's decoder

    Parameters
    -------------
    bvs: np.ndarray
        n_samples x n_alt x n_feat (not z-scaled)
    drivers: np.ndarray
        n_samples x n_drivers input list of driving parameters (not z-scaled).
    driver_names: list
        list of names of driving parameters
    model: str, optional
        name of model to use
    bv_type: str. optional
        name of the type of background variables to use (lidar or radar)

    Returns
    -------------
    samples: np.ndarray
        n_waves x n_alt map found in transformer
        altitude.
    """
    n_batch = drivers.shape[0]

    # verify the correct shapes for the inputs
    if drivers.shape != (n_batch, len(driver_names)):
        raise ValueError(f"driver and driver_names must have the "
                         f"same length ({drivers.shape[-1]} != {len(driver_names)}")

    # transform inputs
    driver_feat = trans.scale_driver(drivers, driver_names=const.driver_names)
    bv_feat, valid_mask = trans.scale_bv(bvs, bv_type)

    driver_feat = torch.tensor(driver_feat, dtype=torch.float)
    bv_feat = torch.tensor(bv_feat, dtype=torch.float)

    bv_feat, alt_mask = mask_altitude(bv_feat)
    hfp_gan, _ = persist.recreate(name=model)
    hfp_gan.eval()
    transformer = hfp_gan.transformer
    src_key_mask = ~alt_mask

    with torch.no_grad():
        bv_shift = torch.roll(bv_feat.transpose(0, 1) @ transformer.bv_emb, 1, dims=0)
        bv_shift[0, :, :] = transformer.bv_start_token.unsqueeze(1)
        src = (driver_feat.unsqueeze(0) @ transformer.dr_emb + bv_shift + transformer.bv_position.unsqueeze(1))
        tgt = (driver_feat.unsqueeze(0) @ transformer.dr_emb + transformer.hfp_start_token.unsqueeze(1))
        memory = transformer.encoder(src, mask=transformer.src_mask,
                                     src_key_padding_mask=src_key_mask.to(transformer.src_mask.dtype))
        attn_map = get_decoder_attn(transformer, tgt, memory, src_key_mask.to(transformer.src_mask.dtype))

    attn_map = torch.mean(attn_map, dim=0).detach().numpy()
    return attn_map


def collect_bv_attn_map(drivers: np.ndarray,
                        bvs: np.ndarray,
                        driver_names: List[str] = const.driver_names,
                        model: str = 'bv_gan',
                        bv_type: str = 'radar'):
    """
    function to collect attention map from weights in pre-trained model

    Parameters
    -------------
    bvs: np.ndarray
        n_samples x n_alt x n_feat (not z-scaled)
    drivers: np.ndarray
        n_samples x n_drivers input list of driving parameters (not z-scaled).
    driver_names: list
        list of names of driving parameters
    model: str, optional
        name of model to use
    bv_type: str. optional
        name of the type of background variables to use (lidar or radar)

    Returns
    -------------
    samples: np.ndarray
        n_alt x n_alt atten map found in transformer
        altitude.
    """
    n_batch = drivers.shape[0]

    # verify the correct shapes for the inputs
    if drivers.shape != (n_batch, len(driver_names)):
        raise ValueError(f"driver and driver_names must have the "
                         f"same length ({drivers.shape[-1]} != {len(driver_names)}")

    # transform inputs
    driver_feat = trans.scale_driver(drivers, driver_names=const.driver_names)
    bv_feat, valid_mask = trans.scale_bv(bvs, bv_type)

    driver_feat = torch.tensor(driver_feat, dtype=torch.float)
    bv_feat = torch.tensor(bv_feat, dtype=torch.float)

    bv_feat, alt_mask = mask_altitude(bv_feat)
    gan, _ = persist.recreate(name=model)
    gan.eval()
    transformer = gan.transformer
    src_key_mask = ~alt_mask

    with torch.no_grad():
        bv_shift = torch.roll(bv_feat.transpose(0, 1) @ transformer.bv_emb, 1, dims=0)
        bv_shift[0, :, :] = transformer.start_token.unsqueeze(1)
        src = (driver_feat.unsqueeze(0) @ transformer.dr_emb + bv_shift + transformer.position.unsqueeze(1))
        encoder = transformer.encoder

        attn_output_weights = get_encoder_attn(encoder.layers, src,
                                               transformer.src_mask,
                                               src_key_mask.to(transformer.src_mask.dtype))

    attn_map = torch.mean(attn_output_weights, dim=0)
    return attn_map.detach().numpy()



import unittest
import abcgan.constants as const
import numpy as np
import os
import itertools
import h5py
import abcgan.transforms as trans
from abcgan.interface import estimate_drivers, load_h5_data, generate_multi_hfp, load_wtec_h5
from abcgan.interface import generate_wtec, generate_multi_bv, average_wtec, stack_bvs, stack_drivers
from abcgan.interface import generate_bvs, generate_hfps, generate_multi_wtec, discriminate

dir_path = os.path.dirname(os.path.realpath(__file__))
fname = os.path.join(dir_path, "..", "tutorials", "tutorial_isr.h5")
wtec_fname = os.path.join(dir_path, "..", "tutorials", f"wtec_tutorial.h5")

with h5py.File(fname, 'r') as hf:
    nSamples = hf['Drivers'][const.driver_names[0]].shape[0]


def fake_drivers(n, n_drivers=const.n_driver):
    return np.exp(np.random.normal(size=(n, n_drivers)))


def fake_bvs(n):
    bvs = [np.random.uniform(low=const.bv_thresholds[i, 0],
                             high=const.bv_thresholds[i, 1],
                             size=(n, const.max_alt)) for i in range(const.n_bv)]
    return np.stack(bvs, axis=-1)


def fake_hfp(n):
    hfps = [np.random.uniform(low=const.hfp_thresholds[i, 0],
                              high=const.hfp_thresholds[i, 1],
                              size=n) for i in range(const.n_hfp)]
    return np.stack(hfps, axis=-1)[:, None, :]


def fake_lidar_bvs(n):
    return np.exp(np.random.normal(size=(n, const.max_alt_lidar, const.n_lidar_bv)))


def fake_wtec(n, tid_type=const.wtec_default_tid_type):
    wtec = np.zeros((n, const.n_wtec))
    for i in range(wtec.shape[-1]):
        wtec[:, i] = np.random.uniform(low=const.wtec_dict[tid_type]['meas_ranges'][i, 0],
                                       high=const.wtec_dict[tid_type]['meas_ranges'][i, 1],
                                       size=n)
    return wtec


def get_combinations(tid_type, n_tests):
    combinations = []
    for i in range(1, len(const.wtec_sites.keys())):
        combinations.extend(list(itertools.combinations([loc for loc in list(const.wtec_sites.keys())
                                                         if f'{tid_type}_{loc}' in const.wtec_dataset_names],
                                                        i)))
    combinations = [list(c) for c in combinations]
    if len(combinations) > n_tests:
        rand_indexes = list(np.random.choice(np.arange(len(combinations)), replace=False, size=(n_tests,)).astype(int))
        combinations = [c for i, c in enumerate(combinations) if i in rand_indexes]
    return combinations


class TestInterface(unittest.TestCase):

    def test_driver_estimation(self):
        drivers = fake_drivers(5)
        est_drs = estimate_drivers(drivers)
        self.assertEqual(est_drs.shape, drivers.shape)

    def test_generator(self):
        drivers = fake_drivers(5)
        bvs = generate_bvs(drivers, verbose=0)
        self.assertEqual(bvs.shape, (len(drivers), const.max_alt, const.n_bv))

    def test_generator_bv_inputs(self):
        top_alt = np.random.randint(low=0, high=const.max_alt)
        drivers = fake_drivers(5)
        bvs = fake_bvs(5)[:, :top_alt, :]
        G_bvs = generate_bvs(drivers, bv_measurements=bvs, verbose=0)
        self.assertEqual(G_bvs.shape[0], drivers.shape[0])
        self.assertTrue((bvs == G_bvs[:, :top_alt, :]).all())
        self.assertEqual(G_bvs.shape, (len(bvs), const.max_alt, const.n_bv))

    def test_hfp_generator(self):
        drivers = fake_drivers(5)
        bvs = fake_bvs(5)
        hfps, b = generate_hfps(drivers, bvs, verbose=0)
        self.assertEqual(hfps.shape, (drivers.shape[0], const.n_waves, const.n_hfp))
        self.assertEqual(b.shape, (drivers.shape[0], ))

    def test_wtec_generator(self):
        for ds in const.wtec_dataset_names:
            tid_type, loc = ds[:6], ds[7:]
            drivers = fake_drivers(5, const.n_wtec_dr)
            wtecs = generate_wtec(drivers, location=loc, tid_type=tid_type, verbose=0)
            self.assertEqual(wtecs.shape, (drivers.shape[0], const.n_wtec))
            self.assertTrue(not (np.isnan(wtecs).any()))
            wtec_feats = generate_wtec(drivers, location=loc, tid_type=tid_type,
                                       return_z_scale=True, verbose=0)
            self.assertEqual(wtec_feats.shape, (drivers.shape[0], const.n_wtec))
            self.assertTrue(not (np.isnan(wtec_feats).any()))

    def test_bv_generator_missing_alts(self):
        top_alt = np.random.randint(low=1, high=const.max_alt)
        bvs = fake_bvs(5)[:, :top_alt, :]
        drivers = fake_drivers(5)
        G_bvs = generate_bvs(drivers, bv_measurements=bvs, verbose=0)
        self.assertEqual(G_bvs.shape, (drivers.shape[0], const.max_alt, const.n_bv))
        self.assertTrue(~(np.isnan(G_bvs).any()))
        self.assertTrue((bvs == G_bvs[:, :top_alt, :]).all())

    def test_hfp_generator_missing_alts(self):
        top_alt = np.random.randint(low=1, high=const.max_alt)
        bvs = fake_bvs(5)[:, :top_alt, :]
        tmp_bvs = bvs.copy()
        drivers = fake_drivers(5)
        G_hfps, b = generate_hfps(drivers, bv_measurements=tmp_bvs, verbose=0)
        self.assertTrue((bvs == tmp_bvs).all())
        self.assertTrue(~(np.isnan(G_hfps).any()))
        self.assertEqual(G_hfps.shape, (drivers.shape[0], const.n_waves, const.n_hfp))
        self.assertEqual(b.shape, (drivers.shape[0], ))

    def test_bv_gen_multi_inputs(self):
        n_repeat = 5
        drivers = fake_drivers(5)
        G_bvs = generate_multi_bv(drivers, n_repeat=n_repeat, verbose=0)
        self.assertEqual(G_bvs.shape, (drivers.shape[0], n_repeat, const.max_alt, const.n_bv))
        self.assertTrue(~(np.isnan(G_bvs).any()))

    def test_gen_multi_missing_alts(self):
        n_repeat = 5
        top_alt = np.random.randint(low=1, high=const.max_alt)
        drivers = fake_drivers(5)
        bvs = fake_bvs(5)[:, :top_alt, :]
        G_bvs = generate_multi_bv(drivers, bvs=bvs, n_repeat=n_repeat, verbose=0)
        self.assertEqual(G_bvs.shape, (drivers.shape[0], n_repeat, const.max_alt, const.n_bv))
        self.assertTrue((bvs == G_bvs[:, 0, :top_alt, :]).all())
        self.assertTrue(~(np.isnan(G_bvs).any()))

    def test_hfp_gen_multi(self):
        n_repeat = 5
        drivers = fake_drivers(5)
        bvs = fake_bvs(5)
        G_hfps, G_bs = generate_multi_hfp(drivers, bvs=bvs, n_repeat=n_repeat, verbose=0)
        self.assertEqual(G_hfps.shape, (drivers.shape[0], n_repeat, const.n_waves, const.n_hfp))
        self.assertEqual(G_bs.shape, (drivers.shape[0], n_repeat, ))
        self.assertTrue(~(np.isnan(G_hfps).any()))
        self.assertTrue(~(np.isnan(G_bs).any()))

    def test_hfp_gen_multi_missing_alts(self):
        n_repeat = 5
        top_alt = np.random.randint(low=1, high=const.max_alt)
        drivers = fake_drivers(5)
        bvs = fake_bvs(5)[:, :top_alt, :]
        G_hfps, G_bs = generate_multi_hfp(drivers, bvs=bvs, n_repeat=n_repeat, verbose=0)
        self.assertEqual(G_hfps.shape, (drivers.shape[0], n_repeat, const.n_waves, const.n_hfp))
        self.assertEqual(G_bs.shape, (drivers.shape[0], n_repeat, ))
        self.assertTrue(~(np.isnan(G_hfps).any()))
        self.assertTrue(~(np.isnan(G_bs).any()))

    def test_wtec_gen_multi(self):
        n_repeat = 5
        drivers = fake_drivers(5, const.n_wtec_dr)
        for ds in const.wtec_dataset_names:
            tid_type, loc = ds[:6], ds[7:]
            G_wtec = generate_multi_wtec(drivers, tid_type=tid_type, location=loc,
                                         n_repeat=n_repeat, verbose=0)
            self.assertEqual(G_wtec.shape, (drivers.shape[0], n_repeat, const.n_wtec))

    def test_hfp_discriminator(self):
        drivers = fake_drivers(5)
        bvs = fake_bvs(5)
        hfps = fake_hfp(5)
        bv_scores, hfp_scores = discriminate(drivers, bvs, hfps=hfps)
        self.assertEqual(bv_scores.shape, (drivers.shape[0], bvs.shape[1]))
        self.assertEqual(hfp_scores.shape, (drivers.shape[0], hfps.shape[1]))

    def test_lidar_generator(self):
        drivers = fake_drivers(5)
        bvs = generate_bvs(drivers, n_alt=const.max_alt_lidar,
                           bv_model='bv_lidar_gan', bv_type='lidar',
                           verbose=0)
        self.assertEqual(bvs.shape, (len(drivers), const.max_alt_lidar, const.n_lidar_bv))

    def test_lidar_discriminator(self):
        drivers = fake_drivers(5)
        bvs = fake_lidar_bvs(5)
        disc = discriminate(drivers, bvs, bv_model='bv_lidar_gan', bv_type='lidar')
        self.assertEqual(disc.shape, (drivers.shape[0], bvs.shape[1]))

    def test_discriminator(self):
        drivers = fake_drivers(5)
        bvs = fake_bvs(5)
        disc = discriminate(drivers, bvs, bv_model='bv_gan')
        self.assertEqual(disc.shape, (drivers.shape[0], bvs.shape[1]))

    def test_stack_drivers(self):
        with h5py.File(fname, 'r') as hf:
            drivers = stack_drivers(hf['Drivers'])
        self.assertTrue(isinstance(drivers, np.ndarray))
        self.assertEqual(drivers.shape, (nSamples, const.n_driver))

    def test_stack_lidar_bvs(self):
        with h5py.File(fname, 'r') as hf:
            bvs = stack_bvs(hf['BackgroundValues'], bv_type='lidar')
        self.assertTrue(isinstance(bvs, np.ndarray))
        self.assertEqual(bvs.shape,
                         (nSamples, const.max_alt_lidar, const.n_lidar_bv))

    def test_stack_bvs_radar(self):
        with h5py.File(fname, 'r') as hf:
            bvs = stack_bvs(hf['BackgroundValues'], bv_type='radar')
        self.assertTrue(isinstance(bvs, np.ndarray))
        self.assertEqual(bvs.shape,
                         (nSamples, const.max_alt, const.n_bv))

    def test_load_h5_data_bv(self):
        drivers, bvs, alt_mask, unix_time = load_h5_data(fname)
        n_samples = len(drivers)
        self.assertTrue(len(drivers) == len(bvs) == len(alt_mask) == len(unix_time))
        self.assertEqual(drivers.shape, (n_samples, const.n_driver))
        self.assertEqual(bvs.shape, (n_samples, const.max_alt, const.n_bv))
        self.assertEqual(alt_mask.shape, (n_samples, const.max_alt))
        self.assertEqual(unix_time.shape, (n_samples,))
        n_samples = 500
        drivers, bvs, alt_mask, unix_time = load_h5_data(fname, n_samples=n_samples)
        self.assertTrue(len(drivers) == len(bvs) == len(alt_mask) == len(unix_time))
        self.assertEqual(drivers.shape, (n_samples, const.n_driver))
        self.assertEqual(bvs.shape, (n_samples, const.max_alt, const.n_bv))
        self.assertEqual(alt_mask.shape, (n_samples, const.max_alt))
        self.assertEqual(unix_time.shape, (n_samples,))
        drivers, bvs, alt_mask, hfps, wave_mask, unix_time = load_h5_data(fname, load_hfp=True)
        n_samples = len(drivers)
        self.assertTrue(len(drivers) == len(bvs) == len(alt_mask) == len(unix_time) == len(hfps) == len(wave_mask))
        self.assertEqual(drivers.shape, (n_samples, const.n_driver))
        self.assertEqual(bvs.shape, (n_samples, const.max_alt, const.n_bv))
        self.assertEqual(alt_mask.shape, (n_samples, const.max_alt))
        self.assertEqual(unix_time.shape, (n_samples,))
        n_samples = 500
        drivers, bvs, alt_mask, hfps, wave_mask, unix_time = load_h5_data(fname, n_samples=n_samples, load_hfp=True)
        self.assertTrue(len(drivers) == len(bvs) == len(alt_mask) == len(unix_time) == len(hfps) == len(wave_mask))
        self.assertEqual(drivers.shape, (n_samples, const.n_driver))
        self.assertEqual(bvs.shape, (n_samples, const.max_alt, const.n_bv))
        self.assertEqual(alt_mask.shape, (n_samples, const.max_alt))
        self.assertEqual(unix_time.shape, (n_samples,))

    def test_load_tec_h5_start_random(self):
        combinations = get_combinations(const.wtec_default_tid_type, n_tests=2)
        n_samples = [None, np.random.randint(low=1, high=5000)]
        for locs, nb in zip(combinations, n_samples):
            data = load_wtec_h5(fname=wtec_fname, locations=locs, tid_type=const.wtec_default_tid_type,
                                start_utc='random', n_samples=nb)
            self.assertTrue(list(data.keys()) == locs)
            for loc, res in data.items():
                drivers, wtec, unix_time = res["drivers"], res["wtecs"], res["utc"]
                self.assertTrue(len(drivers) == len(wtec) == len(unix_time))
                self.assertEqual(drivers.shape[-1], const.n_wtec_dr)
                self.assertEqual(wtec.shape[-1], const.n_wtec)

    def test_load_tec_h5(self):
        n_test = 2
        for tid_type in list(const.wtec_dict.keys()):
            combinations = get_combinations(tid_type, n_test)
            for locs in combinations:
                data = load_wtec_h5(fname=wtec_fname, locations=locs, tid_type=tid_type)
                self.assertTrue(list(data.keys()) == locs)
                for loc, res in data.items():
                    drivers, wtec, unix_time = res["drivers"], res["wtecs"], res["utc"]
                    n_samples = len(drivers)
                    self.assertTrue(len(drivers) == len(wtec) == len(unix_time))
                    self.assertEqual(drivers.shape, (n_samples, const.n_wtec_dr))
                    self.assertEqual(wtec.shape, (n_samples, const.n_wtec))
                    self.assertEqual(unix_time.shape, (n_samples,))
        n_samples = 500
        locations = [d[7:] for d in const.wtec_dataset_names if d.find(const.wtec_default_tid_type) == 0
                     and d[7:] in const.wtec_sites.keys()]
        data = load_wtec_h5(fname=wtec_fname, locations=locations,
                            tid_type=const.wtec_default_tid_type,
                            n_samples=n_samples)
        for loc, res in data.items():
            drivers, wtec, unix_time = res["drivers"], res["wtecs"], res["utc"]
            self.assertTrue(len(drivers) == len(wtec) == len(unix_time))
            self.assertEqual(drivers.shape, (n_samples, const.n_wtec_dr))
            self.assertEqual(wtec.shape, (n_samples, const.n_wtec))
            self.assertEqual(unix_time.shape, (n_samples,))

    def test_load_tec_h5_context(self):
        n_test = 2
        n_context = 1
        for tid_type in list(const.wtec_dict.keys()):
            combinations = get_combinations(tid_type, n_test)
            for locs in combinations:
                data = load_wtec_h5(fname=wtec_fname, locations=locs, tid_type=tid_type, n_context=n_context)
                self.assertTrue(list(data.keys()) == locs)
                for loc, res in data.items():
                    drivers, wtec, unix_time, ctx_wtecs = res["drivers"], res["wtecs"], res["utc"], res["ctx_wtecs"]
                    n_samples = len(drivers)
                    self.assertTrue(len(drivers) == len(wtec) == len(unix_time) == len(ctx_wtecs))
                    self.assertEqual(drivers.shape, (n_samples, const.n_wtec_dr))
                    self.assertEqual(wtec.shape, (n_samples, const.n_wtec))
                    self.assertEqual(ctx_wtecs.shape, (n_samples, n_context, const.n_wtec))
                    self.assertEqual(unix_time.shape, (n_samples,))

    def test_wtec_averaging(self):
        n_samples = 50
        data = load_wtec_h5(fname=wtec_fname, locations=const.wtec_default_location,
                            tid_type=const.wtec_default_tid_type,
                            n_samples=n_samples)
        for loc, res in data.items():
            drivers, wtec, unix_time = res["drivers"], res["wtecs"], res["utc"]
            wtec_avg = average_wtec(wtec, tid_type=const.wtec_default_tid_type)
            self.assertEqual(wtec_avg.shape, wtec.shape)
            self.assertTrue(np.isclose((np.nanmean(wtec) - np.nanmean(wtec_avg))/np.nanmean(wtec), 0, atol=1e-2))
            self.assertTrue(np.isclose((np.nanstd(wtec) - np.nanstd(wtec_avg))/np.nanstd(wtec), 0, atol=1e-2))

    def test_zscale_wtec_averaging(self):
        n_samples = 50
        data = load_wtec_h5(fname=wtec_fname, locations=const.wtec_default_location,
                            tid_type=const.wtec_default_tid_type,
                            n_samples=n_samples)
        for loc, res in data.items():
            drivers, wtec, unix_time = res["drivers"], res["wtecs"], res["utc"]
            wtec_feat = trans.scale_wtec(wtec, tid_type=const.wtec_default_tid_type)[0]
            wtec_feat_avg = average_wtec(wtec_feat, tid_type=const.wtec_default_tid_type, z_scale_input=True)
            self.assertEqual(wtec_feat_avg.shape, wtec_feat.shape)
            self.assertTrue(np.isclose((np.nanmean(wtec_feat) - np.nanmean(wtec_feat_avg)), 0, atol=1e-1))
            self.assertTrue(np.isclose((np.nanstd(wtec_feat) - np.nanstd(wtec_feat_avg)), 0, atol=1e-1))

    def test_type(self):
        with h5py.File(fname, 'r') as hf:
            driver_dict = {k: v[()] for k, v in hf['Drivers'].items()}
            bv_dict = {k: v[()] for k, v in
                       hf['BackgroundValues'].items()}
            driver_dict[const.driver_names[0]] = None
            with self.assertRaises(ValueError):
                stack_drivers(driver_dict)
            bv_dict[const.bv_names[0]] = None
            with self.assertRaises(ValueError):
                stack_bvs(bv_dict)

    def test_missing_name(self):
        with h5py.File(fname, 'r') as hf:
            driver_dict = {k: v[()] for k, v in hf['Drivers'].items()}
            bv_dict = {k: v[()] for k, v in
                       hf['BackgroundValues'].items()}
            del driver_dict[const.driver_names[0]]
            with self.assertRaises(KeyError):
                stack_drivers(driver_dict)
            del bv_dict[const.bv_names[0]]
            with self.assertRaises(KeyError):
                stack_bvs(bv_dict)

    def test_wrong_shape(self):
        with h5py.File(fname, 'r') as hf:
            driver_dict = {k: v[()] for k, v in hf['Drivers'].items()}
            bv_dict = {k: v[()] for k, v in
                       hf['BackgroundValues'].items()}
            driver_dict[const.driver_names[0]] = \
                driver_dict[const.driver_names[0]][:10]
            with self.assertRaises(ValueError):
                stack_drivers(driver_dict)
            bv_dict[const.bv_names[0]] = \
                bv_dict[const.bv_names[0]][:5]
            with self.assertRaises(ValueError):
                stack_bvs(bv_dict)


if __name__ == "__main__":
    unittest.main(argv=['first-arg-is-ignored'], exit=False)



import unittest
import numpy as np
import os

from .test_interface import fake_wtec, fake_hfp, fake_bvs
import abcgan.constants as const
from abcgan.interface import generate_bvs, generate_hfps, load_h5_data, load_wtec_h5
from abcgan.evaluate import hellinger_scores_hfp, hellinger_scores_bv,\
    hellinger_scores_wtec, conditional_wtec_scores, get_clusters

dir_path = os.path.dirname(os.path.realpath(__file__))
fname = os.path.join(dir_path, "..", "tutorials", "tutorial_isr.h5")
wtec_fname = os.path.join(dir_path, "..", "tutorials", "wtec_tutorial.h5")


class TestEvaluate(unittest.TestCase):

    def test_bv_hellinger_scoring(self):
        n_samples = 50
        drivers, bvs, alt_mask, unix_time = load_h5_data(fname, n_samples=n_samples)
        top_alt = np.random.randint(low=1, high=const.max_alt)
        G_bvs = generate_bvs(drivers, bvs[:, :top_alt, :], verbose=0)
        dist = hellinger_scores_bv(bvs, G_bvs, alt_mask)
        self.assertEqual(dist.shape, (const.max_alt, const.n_bv))
        self.assertTrue(~(np.isnan(dist).any()))
        G_bvs = generate_bvs(drivers, verbose=0)
        dist = hellinger_scores_bv(bvs, G_bvs, alt_mask, filter_length=0)
        self.assertEqual(dist.shape, (const.max_alt, const.n_bv))
        self.assertTrue(~(np.isnan(dist).any()))
        dist = hellinger_scores_bv(bvs, G_bvs, alt_mask, z_scale=False)
        self.assertEqual(dist.shape, (const.max_alt, const.n_bv))
        self.assertTrue(~(np.isnan(dist).any()))

    def test_hfp_hellinger_scoring(self):
        n_samples = 50
        drivers, bvs, alt_mask, hfps, wave_mask, unix_time = load_h5_data(fname, n_samples=n_samples, load_hfp=True)
        top_alt = np.random.randint(low=1, high=const.max_alt)
        G_hfps, G_bs = generate_hfps(drivers, bvs[:, :top_alt, :], verbose=0)
        Gb_mask = (G_bs < np.random.uniform(size=len(G_bs)))[:, None]
        dist = hellinger_scores_hfp(hfps, G_hfps, r_mask=wave_mask, f_mask=Gb_mask)
        self.assertTrue(~(np.isnan(dist).any()))
        self.assertEqual(dist.shape, (const.n_waves, const.n_hfp))
        dist = hellinger_scores_hfp(hfps, G_hfps, r_mask=wave_mask, f_mask=Gb_mask, filter_length=0)
        self.assertTrue(~(np.isnan(dist).any()))
        self.assertEqual(dist.shape, (const.n_waves, const.n_hfp))
        dist = hellinger_scores_hfp(hfps, G_hfps, r_mask=wave_mask, f_mask=Gb_mask, z_scale=False)
        self.assertTrue(~(np.isnan(dist).any()))
        self.assertEqual(dist.shape, (const.n_waves, const.n_hfp))

    def test_wtec_hellinger_scoring(self):
        n_samples = 50
        for tid_type in const.wtec_dict.keys():
            locations = [d[7:] for d in const.wtec_dataset_names if d.find(tid_type) == 0
                         and d[7:] in const.wtec_sites.keys()]
            data = load_wtec_h5(fname=wtec_fname, locations=locations, tid_type=tid_type, n_samples=n_samples, start_utc='random')
            for loc, res in data.items():
                drivers, wtec, unix_time = res["drivers"], res["wtecs"], res["utc"]
                wtec_fake = fake_wtec(n_samples, tid_type=tid_type)
                dist = hellinger_scores_wtec(wtec, wtec_fake, tid_type=tid_type)
                self.assertEqual(dist.shape, (const.n_wtec,))
                dist = hellinger_scores_wtec(wtec, wtec_fake, tid_type=tid_type, filter_length=0)
                self.assertEqual(dist.shape, (const.n_wtec,))
                dist = hellinger_scores_wtec(wtec, wtec_fake, tid_type=tid_type, use_kde=True)
                self.assertEqual(dist.shape, (const.n_wtec,))
                dist = hellinger_scores_wtec(wtec, wtec_fake, tid_type=tid_type, filter_length=2, z_scale=False)
                self.assertEqual(dist.shape, (const.n_wtec,))

    def test_wtec_conditional_scoring(self):
        n_samples = 100
        n_clusters = 3
        tid_type = const.wtec_default_tid_type
        locations = [d[7:] for d in const.wtec_dataset_names if d.find(tid_type) == 0
                     and d[7:] in const.wtec_sites.keys()]
        data = load_wtec_h5(fname=wtec_fname, locations=locations, tid_type=tid_type, n_samples=n_samples, start_utc='random')
        for loc, res in data.items():
            drivers, wtec, unix_time = res["drivers"], res["wtecs"], res["utc"]
            wtec_fake = fake_wtec(n_samples, tid_type=tid_type)
            dist = conditional_wtec_scores(drivers, wtec, wtec_fake, n_clusters=n_clusters, tid_type=tid_type)
            self.assertEqual(dist.shape, (drivers.shape[-1], n_clusters, const.n_wtec,))
            dist = conditional_wtec_scores(drivers, wtec, wtec_fake, tid_type=tid_type, n_clusters=n_clusters, filter_length=0)
            self.assertEqual(dist.shape, (drivers.shape[-1], n_clusters, const.n_wtec,))
            dist = conditional_wtec_scores(drivers, wtec, wtec_fake, tid_type=tid_type, n_clusters=n_clusters, use_kde=True)
            self.assertEqual(dist.shape, (drivers.shape[-1], n_clusters, const.n_wtec,))
            dist = conditional_wtec_scores(drivers, wtec, wtec_fake, n_clusters=n_clusters,
                                           tid_type=tid_type, filter_length=2, z_scale=True)
            self.assertEqual(dist.shape, (drivers.shape[-1], n_clusters, const.n_wtec,))

    def test_get_clusters(self):
        n_samples = 100
        n_clusters = np.arange(1, 5)
        tid_type = const.wtec_default_tid_type
        locations = const.wtec_default_location
        data = load_wtec_h5(fname=wtec_fname, locations=locations, tid_type=tid_type, n_samples=n_samples, start_utc='random')
        for loc, res in data.items():
            drivers, wtec, unix_time = res["drivers"], res["wtecs"], res["utc"]
            for c in n_clusters:
                clusters = get_clusters(drivers, c)
                self.assertTrue((np.unique(clusters) == np.arange(c)).all())


if __name__ == '__main__':
    unittest.main()



import torch
import numpy as np
from typing import Union
import abcgan.constants as const


def prev_driver_mask(unix_time: np.ndarray):
    """
    Creates a driver mask of samples that have a previous sample
    and a mapping vector to the previous sample.

    Parameters
    ----------------
    unix_time: np.array
        time stamp of driver samples

    Returns
    ----------------
    prev_dr_map: np.array
        vector mapping each sample to its delayed sample
    dr_mask: torch.Tensor
        Mask of valid driver samples that have a delayed sample
    """

    dr_mask = np.zeros(len(unix_time), dtype=bool)
    prev_dr_map = []
    for i in range(len(unix_time)):
        diff = np.where(unix_time[i] - unix_time == const.dr_delay)[0]
        if len(diff) > 0:
            dr_mask[i] = True
            prev_dr_map.append(diff.item())
    prev_dr_map = np.array(prev_dr_map)

    return prev_dr_map, dr_mask


def context_mapping(unix_time: Union[list, np.ndarray],
                    n_context: int = 2,
                    n_padding: int = 1,
                    delay: int = 3600):
    """
    Creates a mask of samples that have a previous sample
    and a mapping vector to the previous sample.

    Parameters
    ----------------
    unix_time: np.array
        time stamp of samples
    n_context: int
        the number of previous samples to use as context
    n_padding: int
        the padding offset to begin collecting context samples
    delay: int
        delay in seconds between previous context samples

    Returns
    ----------------
    ctx_mapping: np.array
        vector mapping each sample to its delayed sample
    ctx_mask: np.array
        Mask of valid driver samples that have a delayed sample
    """

    ctx_mask = np.zeros(len(unix_time), dtype=bool)
    ctx_mapping = []
    delays = np.array([delay*(i + n_padding) for i in range(1, n_context + 1)])[::-1]
    for i in range(n_context + n_padding, len(unix_time)):
        c_start = i - n_context - n_padding
        c_stop = i - n_padding
        if (unix_time[i] - unix_time[c_start: c_stop] == delays).all():
            ctx_mapping.append(np.array([i - c - n_padding - 1 for c in range(n_context)])[::-1])
            ctx_mask[i] = True
        else:
            ctx_mask[i] = False
    ctx_mapping = np.array(ctx_mapping)

    return ctx_mapping, ctx_mask


def mask_altitude(bv_feat: torch.Tensor):
    """
    Creates an altitude mask for nans in bvs.

    Also replaces nans with numbers.

    Parameters
    ----------------
    bv_feat: torch.Tensor
        background variables

    Returns
    ----------------
    bv_feat: torch.Tensor
        bv_feat with nans replaced, done in place but returned for clarity
    alt_mask: torch.Tensor
        Mask that is true for valid altitudes

    Raises
    -----------------
    ValueError:
        If valid values are not contiguous.
    """
    squash = False
    if bv_feat.ndim == 2:
        bv_feat = bv_feat.unsqueeze(0)
        squash = True
    alt_mask = ~torch.isnan(bv_feat).any(-1)
    bv_feat[torch.isnan(bv_feat)] = 0.0
    # verify that alt_mask is valid
    chk_mask = torch.nn.functional.pad(alt_mask, (1, 1), value=False)
    start_cnt = (chk_mask[:, 1:] & ~chk_mask[:, :-1]).sum(-1)
    stop_cnt = (~chk_mask[:, 1:] & chk_mask[:, :-1]).sum(-1)
    valid = (((start_cnt == 1) & (stop_cnt == 1)) |
             ((start_cnt == 0) & (stop_cnt == 0)))
    if not valid.all():
        first = np.where(~valid.numpy())[0]
        raise ValueError(f"Invalid data more than one stop / start. "
                         f"First invalid row is: {first}.")
    if squash:
        bv_feat, alt_mask = bv_feat[0, ...], alt_mask[0, ...]
    return bv_feat, alt_mask



import unittest
import abcgan.constants as const
from abcgan import transforms as trans
import numpy as np


def fake_drivers(n, n_drivers=const.n_driver):
    return np.exp(np.random.normal(size=(n, n_drivers)))


def fake_bvs(n):
    return np.exp(np.random.normal(size=(n, const.max_alt, const.n_bv)))


def fake_hfp(n):
    hfp = np.zeros((n, const.n_waves, const.n_hfp))
    for i in range(const.n_hfp):
        for j in range(const.n_waves):
            hfp[:, j, i] = np.random.uniform(low=const.hfp_thresholds[i, 0],
                                             high=const.hfp_thresholds[i, 1],
                                             size=n)
    hfp[np.random.randint(low=0, high=n), :, :] = np.nan
    return hfp


def fake_wtec(n, tid_type):
    wtec = np.zeros((n, const.n_wtec))
    for i in range(const.n_wtec):
        wtec[:, i] = np.random.uniform(low=const.wtec_dict[tid_type]['meas_ranges'][i, 0],
                                       high=const.wtec_dict[tid_type]['meas_ranges'][i, 1],
                                       size=n)
    return wtec


class TestTransforms(unittest.TestCase):

    def test_scale_wtec(self):
        batch_size = 10
        for tid_type in list(const.wtec_dict.keys()):
            wtec = fake_wtec(batch_size, tid_type=tid_type)
            wtec_feat, valid_mask = trans.scale_wtec(wtec, tid_type=tid_type)
            self.assertEqual(wtec_feat.shape, (wtec.shape[0], const.n_wtec))
            self.assertEqual(valid_mask.shape, (batch_size,))

    def test_get_wtec(self):
        for tid_type in list(const.wtec_dict.keys()):
            wtec = fake_wtec(10, tid_type=tid_type)
            wtec_feat, valid_mask = trans.scale_wtec(wtec, tid_type=tid_type)
            new_wtec = trans.get_wtec(wtec_feat, tid_type=tid_type)
            self.assertEqual(new_wtec.shape, wtec.shape)
            self.assertTrue(np.allclose(wtec, new_wtec, equal_nan=True))

    def test_scale_driver(self):
        drivers = fake_drivers(10)
        driver_feat = trans.scale_driver(drivers)
        self.assertEqual(driver_feat.shape,
                         (drivers.shape[0], const.n_driver_feat))

    def test_scale_wtec_driver(self):
        drivers = fake_drivers(10, const.n_wtec_dr)
        driver_feat = trans.scale_driver(drivers, data_type='wtec')
        self.assertEqual(driver_feat.shape,
                         (drivers.shape[0], const.n_wtec_dr_feat))

    def test_scale_bv(self):
        batch_size = 10
        bvs = fake_bvs(batch_size)
        bv_feat, valid_mask = trans.scale_bv(bvs)
        self.assertEqual(bv_feat.shape,
                         (bvs.shape[0], bvs.shape[1],
                          const.n_bv_feat))
        self.assertEqual(valid_mask.shape, (batch_size,))

    def test_scale_hfp(self):
        batch_size = 10
        hfp = fake_hfp(batch_size)
        hfp_feat, valid_mask = trans.scale_hfp(hfp)
        self.assertEqual(hfp_feat.shape,
                         (hfp.shape[0], const.n_waves, const.n_hfp))
        self.assertEqual(valid_mask.shape, (batch_size,))

    def test_get_driver(self):
        drivers = fake_drivers(10)
        driver_feat = trans.scale_driver(drivers)
        new_drivers = trans.get_driver(driver_feat)
        self.assertEqual(drivers.shape, new_drivers.shape)
        self.assertTrue(np.allclose(drivers, new_drivers))

    def test_get_wtec_driver(self):
        drivers = fake_drivers(10, const.n_wtec_dr)
        driver_feat = trans.scale_driver(drivers, data_type='wtec')
        new_drivers = trans.get_driver(driver_feat, data_type='wtec')
        self.assertEqual(drivers.shape, new_drivers.shape)
        self.assertTrue(np.allclose(drivers, new_drivers))

    def test_get_driver_subset(self):
        driver_names = const.driver_names[4:8]
        drivers = fake_drivers(10)[:, :len(driver_names)]
        driver_feat = trans.scale_driver(drivers, driver_names=driver_names)
        new_drivers = trans.get_driver(driver_feat, driver_names=driver_names)
        self.assertEqual(drivers.shape, new_drivers.shape)
        self.assertTrue(np.allclose(drivers, new_drivers))

    def test_get_hfp(self):
        hfp = fake_hfp(10)
        hfp_feat, valid_mask = trans.scale_hfp(hfp)
        new_hfp = trans.get_hfp(hfp_feat)
        self.assertEqual(new_hfp.shape, hfp.shape)
        self.assertTrue(np.allclose(hfp, new_hfp, equal_nan=True))

    def test_get_bv(self):
        bvs = fake_bvs(10)
        bv_feat, valid_mask = trans.scale_bv(bvs)
        new_bvs = trans.get_bv(bv_feat)
        self.assertEqual(new_bvs.shape[1], const.max_alt)
        self.assertTrue(np.allclose(bvs, new_bvs))

    def test_padding(self):
        bvs = fake_bvs(10)
        bvs = bvs[:, :const.max_alt - 5, :]
        bv_feat, valid_mask = trans.scale_bv(bvs)
        self.assertEqual(bv_feat.shape[1], const.max_alt)

    def test_shrink(self):
        bvs = fake_bvs(10)
        bvs = np.concatenate((bvs, bvs), axis=1)
        bv_feat, valid_mask = trans.scale_bv(bvs)
        self.assertEqual(bv_feat.shape[1], const.max_alt)


if __name__ == "__main__":
    unittest.main(argv=['first-arg-is-ignored'], exit=False)



"""
Transforms to and from z-scaled variables.

Uses numpy only (no pytorch)
"""
import numpy as np  # noqa
import abcgan.constants as const
from typing import Union, List


def encoded_driver_names(dr_names: list = const.driver_names):
    """
    Gets list of driver feature names in order of
    Parameters
    ----------
    dr_names : list
        array of driver names.

    Returns
    -------
    driver_feat_names : list
        list of driver feature names
    """
    driver_feat_names = []
    for dn in dr_names:
        if dn in const.cyclic_driver:
            driver_feat_names.append('cos_' + dn)
            driver_feat_names.append('sin_' + dn)
        else:
            driver_feat_names.append(dn)
    return driver_feat_names


def encode(data: np.ndarray, name: str):
    """
    Encode variables, or just add extra dimension

    Parameters
    ----------
    data : np.ndarray
        array of variable values.
    name : str
        name of the variable.

    Returns
    -------
    enc : np.ndarray
        array of encoded variables (with an extra dimension in all cases)
    """
    if name in const.cyclic_driver:
        wrap_val = const.cyclic_driver[name]
        enc = data % wrap_val
        enc = (enc / wrap_val) * 2.0 * np.pi
        enc = np.stack((np.cos(enc), np.sin(enc)), axis=-1)
    else:
        enc = data[..., None]  # add trailing singleton dimension

    return enc


def decode(data: np.ndarray, driver_names: List[str]):
    """
    Encode variables, or just add extra dimension

    Parameters
    ----------
    data : np.ndarray
        array of feature values.
    driver_names: list: str
        list driver names in data
    Returns
    -------
    enc : np.ndarray
        array of encoded variables
    """
    curr = 0
    decs = []
    for name in driver_names:
        if name in const.cyclic_driver:
            wrap_val = const.cyclic_driver[name]
            cval = data[:, curr]
            sval = data[:, curr + 1]
            theta = np.arctan2(sval, cval)
            dec = theta * wrap_val / 2.0 / np.pi
            dec = dec % wrap_val
            decs.append(dec)
            curr += 2
        else:
            decs.append(data[:, curr])
            curr += 1
    return np.stack(decs, axis=-1)


def compute_valid(bvs: np.ndarray,
                  bv_thresholds: np.ndarray = const.bv_thresholds):
    """
    Returns a mask which can be used to get rid of
    invalid background variables samples and outliers

    Parameters
    --------------
    bvs: np.ndarray
        (n_samples x n_waves x n_hfp_feat)
    bv_thresholds: np.ndarray:
        Upper and lower bounds of each bv feature.
    Returns
    --------------
    valid_mask: bool np.ndarray
        (n_samples,)
    """
    # valid only if value within thresholds and if at least one non-zero value
    valid_mask = np.zeros((bvs.shape[0], bv_thresholds.shape[0])) > 0
    for i in range(bv_thresholds.shape[0]):
        valid_mask[:, i] = ~(((bvs[:, :, i] < bv_thresholds[i][0]) |
                             (bvs[:, :, i] > bv_thresholds[i][1])).any(-1) |
                             ((np.isnan(bvs[:, :, i])).all(-1)) |
                             (np.isnan(bvs[:, 0, i])))
    # valid only if every altitude is valid
    valid_mask = valid_mask.all(-1)
    return valid_mask


def compute_valid_hfp(hfps: np.ndarray,
                      hfp_thresholds: np.ndarray = const.hfp_thresholds):
    """
    Returns a mask which can be used to get rid of
    invalid hfp waves and outliers

    Parameters
    --------------
    hfps: np.ndarray
        (n_samples x n_waves x n_hfp_feat)
    hfp_thresholds: np.ndarray:
        Upper and lower bounds of each hfp feature.
    Returns
    --------------
    valid_mask: bool np.ndarray
        (n_samples,)
    """

    # valid only if value within thresholds
    valid_mask = np.zeros((hfps.shape[0], hfp_thresholds.shape[0])) > 0

    for i in range(hfps.shape[1]):
        valid_mask[:, i] = ~(((hfps[:, :, i] < hfp_thresholds[i][0]) |
                             (hfps[:, :, i] > hfp_thresholds[i][1])).any(-1))
    # valid only if every altitude is valid
    valid_mask = valid_mask.any(-1)
    return valid_mask


def compute_valid_wtec(wtec: np.ndarray,
                       wtec_thresholds: Union[np.ndarray, None] = None,
                       spike_threshold: int = 250):
    """
    Returns a mask which can be used to get rid of
    invalid tec waves and outliers

    Parameters
    --------------
    wtec: np.ndarray
        (n_samples x n_waves x n_wtec)
    wtec_thresholds: np.ndarray:
        Upper and lower bounds of each wtec variable.
    spike_threshold: np.ndarray:
        Max unique values count filter
    Returns
    --------------
    valid_mask: bool np.ndarray
        (n_samples,)
    """
    if wtec_thresholds is None:
        wtec_thresholds = np.stack((np.ones(wtec.shape[-1]) * -np.inf,
                                    np.ones(wtec.shape[-1]) * np.inf), axis=-1)

    # valid only if value within thresholds
    valid_mask = np.ones((wtec.shape[0], wtec.shape[1]), dtype=bool)

    for i in range(wtec.shape[1]):
        valid_mask[:, i] &= ~(((wtec[..., i] < wtec_thresholds[i, 0]) |
                               (wtec[..., i] > wtec_thresholds[i, 1])))
        unique, unique_counts = np.unique(wtec[:, i], return_counts=True)
        filter_values = unique[unique_counts > spike_threshold]
        for j in range(len(filter_values)):
            valid_mask[:, i] &= ~(wtec[:, i] == filter_values[j])
        valid_mask[:, i] &= np.isfinite(wtec[..., i])
    # valid only if every altitude is valid
    valid_mask = valid_mask.all(-1)
    return valid_mask


def scale_driver(drivers: np.ndarray,
                 driver_names: list = None,
                 data_type: str = 'isr'):
    """
    Return a scaled version of the drivers.

    Parameters
    --------------
    drivers: np.ndarray
        n_samples x n_driver
    driver_names: list: str
        list of driver names
    data_type: list: str
        either 'isr' for bv and hfps or 'wtec' for gnss data
    Returns
    --------------
    driver_feat: np.ndarray
        n_samples x n_driver_feat
    """
    if data_type == 'wtec':
        driver_feat_names = const.wtec_dr_feat_names
        mu = const.wtec_driver_mu
        sigma = const.wtec_driver_sigma
        if driver_names is None:
            driver_names = const.wtec_dr_names
    else:
        driver_feat_names = const.driver_feat_names
        mu = const.driver_mu
        sigma = const.driver_sigma
        if driver_names is None:
            driver_names = const.driver_names
    driver_feat = np.hstack([
        encode(drivers[:, i], n)
        for i, n in enumerate(driver_names[:drivers.shape[1]])
    ])
    dr_idxs = np.hstack([const.dr_feat_map[name] for name in driver_names])
    log_idxs = [i for (i, d) in enumerate(dr_idxs) if driver_feat_names[d] in const.log_driver_feats]

    driver_feat[:, log_idxs] = np.log(1 + driver_feat[:, log_idxs])
    driver_feat = (driver_feat - mu[dr_idxs]) / sigma[dr_idxs]
    return driver_feat


def get_driver(driver_feat: np.ndarray,
               driver_names: Union[List[str], None] = None,
               data_type: str = 'isr'):
    """
    Invert featurization to recover driving parameters.

    Parameters
    --------------
    driver_feat: np.ndarray
        n_samples x n_driver_feat
    driver_names: list: str
        list driver names in driver_feat
    data_type: list: str
        either 'isr' for bv and hfps or 'wtec' for gnss data
    Returns
    --------------
    original driver: np.ndarray
        n_samples x n_driver
    """
    if data_type == 'wtec':
        driver_feat_names = const.wtec_dr_feat_names
        mu = const.wtec_driver_mu
        sigma = const.wtec_driver_sigma
        if driver_names is None:
            driver_names = const.wtec_dr_names
    else:
        driver_feat_names = const.driver_feat_names
        mu = const.driver_mu
        sigma = const.driver_sigma
        if driver_names is None:
            driver_names = const.driver_names

    dr_idxs = np.hstack([const.dr_feat_map[name] for name in driver_names])
    log_idxs = [i for (i, d) in enumerate(dr_idxs) if driver_feat_names[d] in const.log_driver_feats]

    drivers = sigma[dr_idxs] * driver_feat + mu[dr_idxs]
    drivers[:, log_idxs] = np.exp(drivers[:, log_idxs]) - 1.0
    drivers = decode(drivers, driver_names)
    return drivers


def scale_hfp(hfps: np.ndarray):
    """
    Return a scaled version of the hfps.

    Parameters
    --------------
    hfps: np.ndarray
        n_samples x n_waves x n_hfp

    Returns
    --------------
    hfp_feat: np.ndarray
        n_samples x n_waves x n_hfp_feat
    """
    # Don't compute valid mask if no bvs are available
    if hfps.shape[1] > 0:
        valid_mask = compute_valid_hfp(hfps)
    else:
        valid_mask = np.ones(hfps.shape[0], dtype=bool)

    hfp_feat = hfps.copy()
    hfp_feat[:, :, const.invert_hfp] *= -1
    hfp_feat[:, :, const.log_hfp] = np.log(1 + hfp_feat[:, :, const.log_hfp])
    hfp_feat = (hfp_feat - const.hfp_mu) / const.hfp_sigma
    return hfp_feat, valid_mask


def get_hfp(hfp_feat: np.ndarray):
    """
    Invert featurization to recover hfp.

    Parameters
    --------------
    hfp_feat: np.ndarray
        n_samples x n_hfp_feat

    Returns
    --------------
    hfps: np.ndarray
        n_samples x n_hfp
    """
    hfps = const.hfp_sigma * hfp_feat + const.hfp_mu
    hfps[:, :, const.log_hfp] = np.exp(hfps[:, :, const.log_hfp]) - 1.0
    hfps[:, :, const.invert_hfp] *= -1
    return hfps


def scale_bv(bvs: np.ndarray, bv_type: str = 'radar'):
    """
    Return a scaled version of the drivers.

    Parameters
    --------------
    bvs: np.ndarray
        n_samples x n_bv
    bv_type: str
        string specifying weather to scale

    Returns
    --------------
    bv_feat: np.ndarray
        n_samples x n_bv_feat
    """
    if bv_type == 'lidar':
        thresholds = const.lidar_thresholds
        bv_mu = const.lidar_bv_mu
        bv_sigma = const.lidar_bv_sigma
        max_alt = const.max_alt_lidar
    else:
        thresholds = const.bv_thresholds
        bv_mu = const.bv_mu
        bv_sigma = const.bv_sigma
        max_alt = const.max_alt

    # Don't compute valid mask if no bvs are available
    if bvs.shape[1] > 0:
        valid_mask = compute_valid(bvs, thresholds)
    else:
        valid_mask = np.ones(bvs.shape[0], dtype=bool)

    if bv_type == 'lidar':
        bv_feat = np.log(1 + bvs)
    else:
        bv_feat = bvs.copy()
        bv_feat[:, :, const.log_bvs] = np.log(1 + bv_feat[:, :,  const.log_bvs])
    bv_feat = (bv_feat - bv_mu) / bv_sigma
    # pad bvs to max_alt if needed
    if bv_feat.shape[1] < max_alt:
        pad_alt = max_alt - bv_feat.shape[1]
        bv_feat = np.pad(bv_feat,
                         ((0, 0), (0, pad_alt), (0, 0)),
                         constant_values=np.nan)
    elif bv_feat.shape[1] > max_alt:
        bv_feat = bv_feat[:, :max_alt, :]
    return bv_feat, valid_mask


def get_bv(bv_feat: np.ndarray, bv_type: str = 'radar'):
    """
    Invert featurization to recover bvs.

    Parameters
    --------------
    bv_feat: np.ndarray
        n_samples x n_bv_feat
    bv_type: str
        radar or lidar bvs

    Returns
    --------------
    scaled_feat: np.ndarray
        n_samples x n_bv
    """
    if bv_type == 'lidar':
        bvs = const.lidar_bv_sigma * bv_feat + const.lidar_bv_mu
        bvs = np.exp(bvs) - 1.0
    else:
        bvs = const.bv_sigma * bv_feat + const.bv_mu
        bvs[:, :, const.log_bvs] = np.exp(bvs[:, :, const.log_bvs]) - 1.0
    return bvs


def scale_wtec(wtec: np.ndarray,
               tid_type: str = 'SSTIDs'):
    """
    Return a scaled version of the tec waves.

    Parameters
    --------------
    wtec: np.ndarray
        n_samples x n_wtec_waves x n_wtec
    tid_type: str
        specify dataset type for z-scaling

    Returns
    --------------
    wtec_feat: np.ndarray
        n_samples x n_wtec_waves x n_wtec_feat
    valid_mask: np.ndarray
        n_samples x 1
    """
    # Don't compute valid mask if no bvs are available
    if wtec.shape[1] > 0:
        valid_mask = compute_valid_wtec(wtec,
                                        wtec_thresholds=const.wtec_dict[tid_type]["thresholds"],
                                        spike_threshold=1000)
    else:
        valid_mask = np.ones(wtec.shape[0], dtype=bool)

    if tid_type in const.wtec_dict.keys():
        mu = const.wtec_dict[tid_type]['mu']
        std = const.wtec_dict[tid_type]['sigma']
    else:
        raise ValueError(f"{tid_type} is invalid. Plz choose from:"
                         f" {list(const.wtec_dict.keys())}")

    wtec_feat = (wtec - mu) / std
    return wtec_feat, valid_mask


def get_wtec(wtec_feat: np.ndarray,
             tid_type: str = "SSTIDs"):
    """
    Invert featurization to recover tec waves.

    Parameters
    --------------
    wtec_feat: np.ndarray
        n_samples x n_wtec_feat
    tid_type: str
        specify dataset type for z-scaling

    Returns
    --------------
    wtec: np.ndarray
        n_samples x n_wtec
    """
    if tid_type in const.wtec_dict.keys():
        mu = const.wtec_dict[tid_type]['mu']
        std = const.wtec_dict[tid_type]['sigma']
    else:
        raise ValueError(f"{tid_type} is invalid. Plz choose from:"
                         f" {list(const.wtec_dict.keys())}")

    wtec = std * wtec_feat + mu
    return wtec



import abcgan.constants as const
from torch.nn.init import xavier_uniform_, xavier_normal_
import torch.nn as nn
import torch


class WTEC_Generator(nn.Module):
    """
    Generator Class

    Parameters
    -----------
    n_layers: int
        number of MLP layers
    latent_dim: int
        the dimension of the input latent vector
    feat_dim: int
        the dimension of the output features
    hidden_dim: int
        the inner dimension, a scalar
    """

    def __init__(self,
                 n_layers: int = 2,
                 latent_dim: int = 16,
                 feat_dim: int = const.n_wtec,
                 cond_dim: int = const.n_wtec_dr_feat,
                 hidden_dim: int = 64):
        super(WTEC_Generator, self).__init__()

        input_dim = cond_dim + latent_dim
        self.input_args = {
            'n_layers': n_layers,
            'latent_dim': latent_dim,
            'cond_dim': cond_dim,
            'feat_dim': feat_dim,
            'hidden_dim': hidden_dim,
        }
        # Build the neural network
        layers = [nn.Linear(input_dim, hidden_dim),
                  nn.ReLU(inplace=True)]
        for i in range(n_layers - 1):
            layers.append(nn.Linear(hidden_dim, hidden_dim))
            layers.append(nn.ReLU(inplace=True))
        layers.append(nn.Linear(hidden_dim, feat_dim))
        self.gen_model = nn.Sequential(*layers)

        self.n_layers = n_layers
        self.latent_dim = latent_dim
        self.feat_dim = feat_dim
        self.input_dim = input_dim
        self.cond_dim = cond_dim
        self.hidden_dim = hidden_dim
        self._reset_parameters()

    def forward(self, driver, noise=None):
        """
        forward pass of the generator for generating TEC waves:
        Given input drivers and noise tensor,
        returns generated TEC wave parameters for each time sample.
        Parameters
        --------------
        driver: torch.Tensor
            tensor of  driver features from data loader (n_batch, n_wtec_dr_feat)
        noise: torch.Tensor, optional
            a noise tensor with dimensions (n_batch, latent_dim)

        Returns
        --------------
        fake_output: torch.Tensor
            generated tec waves (n_batch x n_wtec_feats)
        """
        if noise is None:
            noise = torch.randn(driver.shape[0],
                                self.latent_dim,
                                dtype=driver.dtype,
                                device=driver.device)
        fake_output = self.gen_model(torch.cat((driver, noise), 1))
        return fake_output

    def _reset_parameters(self):
        """Initiate parameters in the transformer model."""
        with torch.no_grad():
            for p in self.parameters():
                if p.dim() > 1:
                    xavier_uniform_(p)


class WTEC_Critic(nn.Module):
    """
    Critic Class

    Parameters
    -------------
    n_layers: int
        number of layers in MLP
    feat_dim: int
        the dimension of wtec features fitted for the dataset used, a scalar
    hidden_dim: int
        the inner dimension, a scalar
    """

    def __init__(self,
                 n_layers: int = 2,
                 feat_dim: int = const.n_wtec,
                 cond_dim: int = const.n_wtec_dr_feat,
                 hidden_dim: int = 64):
        super(WTEC_Critic, self).__init__()

        input_dim = cond_dim + 1 * feat_dim
        self.input_args = {
            'n_layers': n_layers,
            'feat_dim': feat_dim,
            'cond_dim': cond_dim,
            'hidden_dim': hidden_dim,
        }
        # Build the neural network
        layers = [nn.Linear(input_dim, hidden_dim),
                  nn.ReLU(inplace=True)]
        for i in range(n_layers - 1):
            layers.append(nn.Linear(hidden_dim, hidden_dim))
            layers.append(nn.ReLU(inplace=True))
        layers.append(nn.Linear(hidden_dim, 1))
        self.crit_model = nn.Sequential(*layers)

        self.n_layers = n_layers
        self.feat_dim = feat_dim
        self.input_dim = input_dim
        self.cond_dim = cond_dim
        self.hidden_dim = hidden_dim
        self._reset_parameters()

    def forward(self, drivers, wtec_feats):
        """
        forward pass of the critic for tec wave given an image
        tensor, returns a 1-dimension tensor representing a fake/real
        prediction.

        Parameters
        ---------------
        drivers: torch.Tensor
            tensor of driver features (n_batch, n_wtec_dr_feat)
        wtec_feats: torch.Tensor
            tensor of fake or real tec feats (n_batch, n_wtec_waves, n_wtec_feat)
        """

        cond_feat = torch.cat((drivers, wtec_feats), 1)
        pred = self.crit_model(cond_feat)
        return pred

    def _reset_parameters(self):
        """Initiate parameters in the transformer model."""
        with torch.no_grad():
            for p in self.parameters():
                if p.dim() > 1:
                    xavier_uniform_(p)
